"""
Utilities to parse SVG graphics into fpdf.drawing objects.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/SVG.html>
"""

# pyright: reportUnknownLambdaType=false, reportUnknownArgumentType=false, reportUnknownMemberType=false

import logging
import math
import re
import warnings
from copy import deepcopy
from os import PathLike
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Optional

from fontTools.svgLib.path import (
    parse_path,  # pyright: ignore[reportUnknownVariableType]
)

from .enums import GradientSpreadMethod, GradientUnits, PathPaintRule, StrokeCapStyle

try:
    from defusedxml.ElementTree import fromstring as parse_xml_str
except ImportError:
    warnings.warn(
        "defusedxml could not be imported - fpdf2 will not be able to sanitize SVG images provided"
    )
    # nosemgrep: python.lang.security.use-defused-xml.use-defused-xml
    from xml.etree.ElementTree import fromstring as parse_xml_str  # nosec

from . import html
from .drawing import (
    BoundingBox,
    ClippingPath,
    GradientPaint,
    GraphicsContext,
    GraphicsStyle,
    PaintedPath,
    PathPen,
    Text,
    TextRun,
)
from .drawing_primitives import (
    DeviceGray,
    DeviceRGB,
    Point,
    Transform,
    color_from_hex_string,
    color_from_rgb_string,
    force_nodocument,
)
from .image_datastructures import ImageCache, VectorImageInfo
from .output import stream_content_for_raster_image
from .pattern import shape_linear_gradient, shape_radial_gradient

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element  # nosec

    from .drawing import Renderable
    from .fpdf import FPDF
    from .output import ResourceCatalog


LOGGER = logging.getLogger(__name__)


# https://www.w3.org/TR/SVG/Overview.html

_HANDY_NAMESPACES = {
    "svg": "http://www.w3.org/2000/svg",
    "xlink": "http://www.w3.org/1999/xlink",
}

NUMBER_SPLIT = re.compile(r"(?:\s+,\s+|\s+,|,\s+|\s+|,)")
TRANSFORM_GETTER = re.compile(
    r"(matrix|rotate|scale|scaleX|scaleY|skew|skewX|skewY|translate|translateX|translateY)\(([\d\.,\s+-]+)\)"
)

CSS_COMMENT_RE = re.compile(r"/\*.*?\*/", re.DOTALL)
CSS_BLOCK_RE = re.compile(r"(?s)([^{}]+)\{([^{}]*)\}")


def _normalize_css_value(value: Optional[str]) -> Optional[str]:
    if value is None:
        return None
    if isinstance(value, str):
        stripped = value.strip()
        if not stripped:
            return None
        return stripped
    return value


@force_nodocument
class Percent(float):
    """class to represent percentage values"""


unit_splitter = re.compile(r"\s*(?P<value>[-+]?[\d\.]+)\s*(?P<unit>%|[a-zA-Z]*)")

# none of these are supported right now
# https://www.w3.org/TR/css-values-4/#lengths
relative_length_units = {
    "%",  # (context sensitive, depends on which attribute it is applied to)
    "em",  # (current font size)
    "ex",  # (current font x-height)
    # CSS 3
    "ch",  # (advance measure of 0, U+0030 glyph)
    "rem",  # (font-size of the root element)
    "vw",  # (1% of viewport width)
    "vh",  # (1% of viewport height)
    "vmin",  # (smaller of vw or vh)
    "vmax",  # (larger of vw or vh)
    # CSS 4
    "cap",  # (font cap height)
    "ic",  # (advance measure of fullwidth U+6C34 glyph)
    "lh",  # (line height)
    "rlh",  # (root element line height)
    "vi",  # (1% of viewport size in root element's inline axis)
    "vb",  # (1% of viewport size in root element's block axis)
}

absolute_length_units = {
    "in": 72,  # (inches, 72 pt)
    "cm": 72 / 2.54,  # (centimeters, 72 / 2.54 pt)
    "mm": 72 / 25.4,  # (millimeters 72 / 25.4 pt)
    "pt": 1,  # (pdf canonical unit)
    "pc": 12,  # (pica, 12 pt)
    "px": 0.75,  # (reference pixel unit, 0.75 pt)
    # CSS 3
    "Q": 72 / 101.6,  # (quarter-millimeter, 72 / 101.6 pt)
}

angle_units = {
    "deg": math.tau / 360,
    "grad": math.tau / 400,
    "rad": 1,  # pdf canonical unit
    "turn": math.tau,
}


# in CSS the default length unit is px, but as far as I can tell, for SVG interpreting
# unitless numbers as being expressed in pt is more appropriate. Particularly, the
# scaling we do using viewBox attempts to scale so that 1 svg user unit = 1 pdf pt
# because this results in the output PDF having the correct physical dimensions (i.e. a
# feature with a 1cm size in SVG will actually end up being 1cm in size in the PDF).
@force_nodocument
def resolve_length(length_str: str, default_unit: str = "pt") -> float:
    """Convert a length unit to our canonical length unit, pt."""
    match = unit_splitter.match(length_str)
    if match is None:
        raise ValueError(f"Unable to parse '{length_str}' as a length") from None
    value, unit = match.groups()
    if not unit:
        unit = default_unit

    try:
        return float(value) * absolute_length_units[unit]
    except KeyError:
        if unit in relative_length_units:
            raise ValueError(
                f"{length_str} uses unsupported relative length {unit}"
            ) from None

        raise ValueError(f"{length_str} contains unrecognized unit {unit}") from None


@force_nodocument
def resolve_angle(angle_str: str, default_unit: str = "deg") -> float:
    """Convert an angle value to our canonical angle unit, radians"""
    m = unit_splitter.match(angle_str)
    value: str | float
    if m is not None:
        value, unit = m.groups()
    else:
        value = angle_str
    if not unit:  # pyright: ignore[reportPossiblyUnboundVariable]
        unit = default_unit

    try:
        return float(value) * angle_units[unit]
    except KeyError:
        raise ValueError(f"angle {angle_str} has unknown unit {unit}") from None


@force_nodocument
def xmlns(space: str, name: str) -> str:
    """Create an XML namespace string representation for the given tag name."""
    try:
        space = f"{{{_HANDY_NAMESPACES[space]}}}"
    except KeyError:
        space = ""

    return f"{space}{name}"


@force_nodocument
def xmlns_lookup(space: str, *names: str) -> dict[str, str]:
    """Create a lookup for the given name in the given XML namespace."""

    result = {}
    for name in names:
        result[xmlns(space, name)] = name
        result[name] = name

    return result  # pyright: ignore[reportUnknownVariableType]


@force_nodocument
def without_ns(qualified_tag: str) -> str:
    """Remove the xmlns namespace from a qualified XML tag name"""
    i = qualified_tag.index("}")
    if i >= 0:
        return qualified_tag[i + 1 :]
    return qualified_tag


shape_tags = xmlns_lookup(
    "svg", "rect", "circle", "ellipse", "line", "polyline", "polygon"
)


@force_nodocument
def svgcolor(colorstr: str) -> DeviceRGB | DeviceGray:
    try:
        colorstr = html.COLOR_DICT[colorstr]
    except KeyError:
        pass

    if colorstr.startswith("#"):
        return color_from_hex_string(colorstr)

    if colorstr.startswith("rgb"):
        return color_from_rgb_string(colorstr)

    raise ValueError(f"unsupported color specification {colorstr}")


@force_nodocument
def convert_stroke_width(incoming: str) -> Optional[float]:
    val = resolve_length(incoming)
    if val < 0:
        raise ValueError(f"stroke width {incoming} cannot be negative")
    if val == 0:
        return None
    return val


@force_nodocument
def convert_miterlimit(incoming: str) -> float:
    val = float(incoming)
    if val < 1.0:
        raise ValueError(f"miter limit {incoming} cannot be less than 1")
    return val


@force_nodocument
def clamp_float(min_val: float, max_val: float) -> Callable[[str], float]:
    def converter(value: str) -> float:
        val = float(value)
        if val < min_val:
            return min_val
        if val > max_val:
            return max_val
        return val

    return converter


@force_nodocument
def inheritable(
    value: str, converter: Callable[[str], Any] = lambda value: value
) -> Any:
    if value in ("inherit", "currentColor"):
        return GraphicsStyle.INHERIT
    return converter(value)


@force_nodocument
def optional(
    value: Optional[str], converter: Callable[[str], Any] = lambda noop: noop
) -> Any:
    # Treat missing/empty/whitespace exactly like "not set"
    if value is None:
        return None
    if isinstance(value, str) and not value.strip():
        return None
    if value == "none":
        return None
    return inheritable(value, converter)


# this is mostly SVG 1.1 stuff. SVG 2 changed some of this and the documentation is much
# harder to assemble into something coherently understandable
svg_attr_map: dict[str, Any] = {
    # https://www.w3.org/TR/SVG11/painting.html#FillProperty
    "fill": lambda colorstr: (
        "fill_color",
        (
            optional(colorstr, svgcolor)
            if not (colorstr and colorstr.startswith("url("))
            else None
        ),
    ),
    # https://www.w3.org/TR/SVG11/painting.html#FillRuleProperty
    "fill-rule": lambda fillrulestr: ("intersection_rule", inheritable(fillrulestr)),
    # https://www.w3.org/TR/SVG11/painting.html#FillOpacityProperty
    "fill-opacity": lambda filopstr: (
        "fill_opacity",
        inheritable(filopstr, clamp_float(0.0, 1.0)),
    ),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeProperty
    "stroke": lambda colorstr: (
        "stroke_color",
        (
            optional(colorstr, svgcolor)
            if not (colorstr and colorstr.startswith("url("))
            else None
        ),
    ),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeWidthProperty
    "stroke-width": lambda valuestr: (
        "stroke_width",
        inheritable(valuestr, convert_stroke_width),
    ),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeDasharrayProperty
    "stroke-dasharray": lambda dasharray: (
        "stroke_dash_pattern",
        optional(
            dasharray,
            lambda da: [float(item) for item in NUMBER_SPLIT.split(da) if item],
        ),
    ),
    # stroke-dashoffset may be a percentage, which we don't support currently
    # https://www.w3.org/TR/SVG11/painting.html#StrokeDashoffsetProperty
    "stroke-dashoffset": lambda dashoff: (
        "stroke_dash_phase",
        inheritable(dashoff, float),
    ),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeLinecapProperty
    "stroke-linecap": lambda capstr: ("stroke_cap_style", inheritable(capstr)),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeLinejoinProperty
    "stroke-linejoin": lambda joinstr: ("stroke_join_style", inheritable(joinstr)),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeMiterlimitProperty
    "stroke-miterlimit": lambda limstr: (
        "stroke_miter_limit",
        inheritable(limstr, convert_miterlimit),
    ),
    # https://www.w3.org/TR/SVG11/painting.html#StrokeOpacityProperty
    "stroke-opacity": lambda stropstr: (
        "stroke_opacity",
        inheritable(stropstr, clamp_float(0.0, 1.0)),
    ),
}


@force_nodocument
def apply_styles(
    stylable: PaintedPath | GraphicsContext,
    svg_element: "Element",
    computed_style: Optional[dict[str, Any]] = None,
) -> None:
    """Apply the known styles from `svg_element` to the pdf path/group `stylable`."""
    if computed_style is not None:
        style = computed_style
    else:
        style = {}
        for key, value in html.parse_css_style(
            svg_element.attrib.get("style", "")
        ).items():
            norm_value = _normalize_css_value(value)
            if norm_value is not None:
                style[key] = norm_value
    stylable.style.auto_close = False

    for attr_name, converter in svg_attr_map.items():
        attr_value = _normalize_css_value(style.get(attr_name))
        if attr_value is None:
            attr_value = _normalize_css_value(svg_element.attrib.get(attr_name))
        if attr_value is not None:
            setattr(stylable.style, *converter(attr_value))

    # handle this separately for now
    opacity_str = _normalize_css_value(style.get("opacity"))
    if opacity_str is None:
        opacity_str = _normalize_css_value(svg_element.attrib.get("opacity"))
    if opacity_str is not None:
        opacity = float(opacity_str)
        stylable.style.fill_opacity = opacity
        stylable.style.stroke_opacity = opacity

    tfstr = svg_element.attrib.get("transform")
    if tfstr:
        stylable.transform = convert_transforms(tfstr)


def _preserve_ws(style_map: dict[str, Any], tag: "Element") -> bool:
    # CSS ‘white-space’ wins; otherwise XML’s xml:space
    ws = (style_map.get("white-space") or "").strip()
    if ws in ("pre", "pre-wrap", "break-spaces"):
        return True
    xml_space = tag.attrib.get("{http://www.w3.org/XML/1998/namespace}space")
    return bool(xml_space == "preserve")


def _collapse_ws(s: Optional[str], preserve: bool = False) -> str:
    """
    Collapse sequences of whitespace characters (spaces, tabs, etc.) to a single space
    In some cases, like with `pre` tags, whitespace should be preserved.
    """
    if s is None:
        return ""
    if preserve:
        return s  # keep all spaces/newlines
    return re.sub(r"\s+", " ", s)


def _svg_font_style_to_emphasis(font_style: Optional[str], font_weight: str) -> str:
    # Map CSS-like values to fpdf "B", "I", "BI", ""
    b = False
    i = False
    if font_weight:
        fw = font_weight.strip().lower()
        # numeric 600+ considered bold in CSS, also "bold", "bolder"
        b = fw in ("bold", "bolder") or (fw.isdigit() and int(fw) >= 600)
    if font_style:
        fs = font_style.strip().lower()
        i = fs in ("italic", "oblique")
    if b and i:
        return "BI"
    if b:
        return "B"
    if i:
        return "I"
    return ""


def _get_attr_or_style(
    tag: "Element", name: str, style_map: Optional[dict[str, Any]] = None
) -> Optional[str]:
    """Return the value for a given attribute, honoring computed style overrides."""
    if style_map is not None:
        value = _normalize_css_value(style_map.get(name))
        if value is not None:
            return value
    if style_map is None:
        inline = html.parse_css_style(tag.attrib.get("style", ""))
        value = _normalize_css_value(inline.get(name))
        if value is not None:
            return value
    return _normalize_css_value(tag.attrib.get(name))


def _parse_font_attrs(
    tag: "Element", style_map: Optional[dict[str, Any]] = None
) -> tuple[Optional[str], str, Optional[float], str]:
    """Returns (font_family or None, font_style_emphasis, font_size_pt or None, text_anchor)"""
    family_value = _get_attr_or_style(tag, "font-family", style_map)
    if family_value:
        cleaned = []
        for item in family_value.split(","):
            stripped = item.strip().replace('"', "").replace("'", "")
            if stripped:
                cleaned.append(stripped)
        family = ",".join(cleaned) if cleaned else None
    else:
        family = None

    # style/weight -> B/I/BI/""
    font_style = _get_attr_or_style(tag, "font-style", style_map) or ""
    font_weight = _get_attr_or_style(tag, "font-weight", style_map) or ""
    emphasis = _svg_font_style_to_emphasis(font_style, font_weight)

    # font-size: default 16px (SVG/CSS); resolve_length defaults to pt; px=0.75pt defined above.
    fs = _get_attr_or_style(tag, "font-size", style_map)
    size_pt = (
        resolve_length(fs) if fs is not None else None
    )  # px handled in absolute_length_units

    # text-anchor: start|middle|end (default start)
    ta = (_get_attr_or_style(tag, "text-anchor", style_map) or "start").strip().lower()
    if ta not in ("start", "middle", "end"):
        ta = "start"

    return family, emphasis, float(size_pt) if size_pt is not None else None, ta


def _parse_xy_delta(
    tag: "Element",
    style_map: Optional[dict[str, Any]] = None,
    font_size: Optional[float] = None,
) -> tuple[float, float, float, float]:
    """x,y (default 0,0) and optional dx,dy (default 0)"""

    # pylint: disable=too-many-return-statements
    def first_number(
        name: str,
        default: float = 0.0,
        resolver: Callable[[str], float] = resolve_length,
    ) -> float:
        attr = None
        if style_map is not None and name in style_map:
            attr = style_map[name]
        elif name in tag.attrib:
            attr = tag.attrib[name]
        if not attr:
            return default
        parts = NUMBER_SPLIT.split(attr.strip())
        if not parts or not parts[0]:
            return default
        token = parts[0]
        match = unit_splitter.fullmatch(token)
        if match:
            value_str = match.group("value")
            unit = (match.group("unit") or "").strip()
            if unit.lower() in {"em", "rem"}:
                if font_size is None:
                    LOGGER.warning(
                        "Ignoring relative %s length on <%s %s> because font size is unknown",
                        unit,
                        tag.tag,
                        name,
                    )
                    return default
                try:
                    return float(value_str) * font_size
                except (TypeError, ValueError):
                    return default
            if unit.lower() == "ex":
                if font_size is None:
                    LOGGER.warning(
                        "Ignoring relative ex length on <%s %s> because font size is unknown",
                        tag.tag,
                        name,
                    )
                    return default
                try:
                    return float(value_str) * font_size * 0.5
                except (TypeError, ValueError):
                    return default
        try:
            return float(resolver(token))
        except ValueError:
            LOGGER.warning(
                "Ignoring unsupported length '%s' on <%s %s>",
                token,
                tag.tag,
                name,
            )
            return default

    x = first_number("x", 0.0)
    y = first_number("y", 0.0)
    dx = first_number("dx", 0.0)
    dy = first_number("dy", 0.0)
    return x, y, dx, dy


def _extract_css_class_styles(css_text: str) -> list[tuple[str, dict[str, Any]]]:
    styles: list[tuple[str, dict[str, Any]]] = []
    if not css_text:
        return styles

    cleaned = CSS_COMMENT_RE.sub("", css_text)
    for selector_block, body in CSS_BLOCK_RE.findall(cleaned):
        declarations = html.parse_css_style(body)
        if not declarations:
            continue
        normalized = {}
        for key, value in declarations.items():
            norm_value = _normalize_css_value(value)
            if norm_value is not None:
                normalized[key] = norm_value
        if not normalized:
            continue
        selectors = [selector.strip() for selector in selector_block.split(",")]
        for selector in selectors:
            if not selector.startswith("."):
                continue
            class_name = selector[1:].split(":", 1)[0].strip()
            if not class_name:
                continue
            styles.append((class_name, dict(normalized)))
    return styles


@force_nodocument
class ShapeBuilder:
    """A namespace within which methods for converting basic shapes can be looked up."""

    @staticmethod
    def new_path(tag: "Element", clipping_path: bool = False) -> PaintedPath:
        """Create a new path with the appropriate styles."""
        path = PaintedPath()
        if clipping_path:
            path = ClippingPath()
        apply_styles(path, tag)
        return path

    @classmethod
    def rect(cls, tag: "Element", clipping_path: bool = False) -> PaintedPath:
        """Convert an SVG <rect> into a PDF path."""
        # svg rect is wound clockwise
        x = resolve_length(tag.attrib.get("x", "0"))
        y = resolve_length(tag.attrib.get("y", "0"))
        width_str = tag.attrib.get("width") or "0"
        if width_str.endswith("%"):
            width: Percent | float = Percent(width_str[:-1])
        else:
            width = resolve_length(width_str)
        height_str = tag.attrib.get("height") or "0"
        if height_str.endswith("%"):
            height: Percent | float = Percent(height_str[:-1])
        else:
            height = resolve_length(height_str)
        rx_str = tag.attrib.get("rx") or "auto"
        ry_str = tag.attrib.get("ry") or "auto"

        if rx_str == "none":
            rx_str = "0"
        if ry_str == "none":
            ry_str = "0"

        rx: float
        ry: float

        if rx_str == ry_str == "auto":
            rx = ry = 0
        elif rx_str == "auto":
            rx = ry = float(ry_str)
        elif ry_str == "auto":
            ry = rx = float(rx_str)
        else:
            rx = float(rx_str)
            ry = float(ry_str)

        if (width < 0) or (height < 0) or (rx < 0) or (ry < 0):
            raise ValueError(f"bad rect {tag}")

        if (width == 0) or (height == 0):
            return PaintedPath()

        if rx > (width / 2):
            rx = width / 2
        if ry > (height / 2):
            ry = height / 2

        path = cls.new_path(tag, clipping_path)
        path.rectangle(x, y, width, height, rx, ry)
        return path

    @classmethod
    def circle(cls, tag: "Element", clipping_path: bool = False) -> PaintedPath:
        """Convert an SVG <circle> into a PDF path."""
        cx = float(tag.attrib.get("cx", 0))
        cy = float(tag.attrib.get("cy", 0))
        r = float(tag.attrib["r"])

        path = cls.new_path(tag, clipping_path)
        path.circle(cx, cy, r)
        return path

    @classmethod
    def ellipse(cls, tag: "Element", clipping_path: bool = False) -> PaintedPath:
        """Convert an SVG <ellipse> into a PDF path."""
        cx = float(tag.attrib.get("cx", 0))
        cy = float(tag.attrib.get("cy", 0))

        rx_str = tag.attrib.get("rx") or "auto"
        ry_str = tag.attrib.get("ry") or "auto"

        path = cls.new_path(tag, clipping_path)

        if (rx_str == ry_str == "auto") or (rx_str == "0") or (ry_str == "0"):
            return path

        if rx_str == "auto":
            rx = ry = float(ry_str)
        elif ry_str == "auto":
            rx = ry = float(rx_str)
        else:
            rx = float(rx_str)
            ry = float(ry_str)

        path.ellipse(cx, cy, rx, ry)
        return path

    @classmethod
    def line(cls, tag: "Element") -> PaintedPath:
        """Convert an SVG <line> into a PDF path."""
        x1 = float(tag.attrib["x1"])
        y1 = float(tag.attrib["y1"])
        x2 = float(tag.attrib["x2"])
        y2 = float(tag.attrib["y2"])

        path = cls.new_path(tag)
        path.move_to(x1, y1)
        path.line_to(x2, y2)
        return path

    @classmethod
    def polyline(cls, tag: "Element") -> PaintedPath:
        """Convert an SVG <polyline> into a PDF path."""
        path = cls.new_path(tag)
        points = "M" + tag.attrib["points"]
        svg_path_converter(path, points)
        return path

    @classmethod
    def polygon(cls, tag: "Element", clipping_path: bool = False) -> PaintedPath:
        """Convert an SVG <polygon> into a PDF path."""
        path = cls.new_path(tag, clipping_path)
        points = "M" + tag.attrib["points"] + "Z"
        svg_path_converter(path, points)
        return path


@force_nodocument
def convert_transforms(tfstr: str) -> Transform:
    """Convert SVG/CSS transform functions into PDF transforms."""

    # SVG 2 uses CSS transforms. SVG 1.1 transforms are slightly different. I'm really
    # not sure if it is worth it to try to support SVG 2 because it is significantly
    # more entangled with The HTML Disaster than SVG 1.1, which makes it astronomically
    # harder to support.
    # https://drafts.csswg.org/css-transforms/#two-d-transform-functions
    parsed = TRANSFORM_GETTER.findall(tfstr)
    # pylint: disable=redefined-loop-name
    transform = Transform.identity()
    for tf_type, args in parsed:
        args = args.strip()
        if tf_type == "matrix":
            a, b, c, d, e, f = tuple(float(n) for n in NUMBER_SPLIT.split(args))
            transform = Transform(a, b, c, d, e, f) @ transform

        elif tf_type == "rotate":
            theta, *about = NUMBER_SPLIT.split(args)
            theta = resolve_angle(theta)
            rotation = Transform.rotation(theta=theta)
            if about:
                # this is an SVG 1.1 feature. SVG 2 uses the transform-origin property.
                # see: https://www.w3.org/TR/SVG11/coords.html#TransformAttribute
                if len(about) == 2:
                    rotation = rotation.about(float(about[0]), float(about[1]))
                else:
                    raise ValueError(
                        f"rotation transform {tf_type}({args}) is malformed"
                    )

            transform = rotation @ transform

        elif tf_type == "scale":
            # if sy is not provided, it takes a value equal to sx
            args = NUMBER_SPLIT.split(args)
            if len(args) == 2:
                sx = float(args[0])
                sy = float(args[1])
            elif len(args) == 1:
                sx = sy = float(args[0])
            else:
                raise ValueError(f"bad scale transform {tfstr}")

            transform = Transform.scaling(x=sx, y=sy) @ transform

        elif tf_type == "scaleX":  # SVG 2
            transform = Transform.scaling(x=float(args), y=1) @ transform

        elif tf_type == "scaleY":  # SVG 2
            transform = Transform.scaling(x=1, y=float(args)) @ transform

        elif tf_type == "skew":  # SVG 2, not the same as skewX@skewY
            # if sy is not provided, it takes a value equal to 0
            args = NUMBER_SPLIT.split(args)
            if len(args) == 2:
                sx = resolve_angle(args[0])
                sy = resolve_angle(args[1])
            elif len(args) == 1:
                sx = resolve_angle(args[0])
                sy = 0
            else:
                raise ValueError(f"bad skew transform {tfstr}")

            transform = Transform.shearing(x=math.tan(sx), y=math.tan(sy)) @ transform

        elif tf_type == "skewX":
            transform = (
                Transform.shearing(x=math.tan(resolve_angle(args)), y=0) @ transform
            )

        elif tf_type == "skewY":
            transform = (
                Transform.shearing(x=0, y=math.tan(resolve_angle(args))) @ transform
            )

        elif tf_type == "translate":
            # if y is not provided, it takes a value equal to 0
            args = NUMBER_SPLIT.split(args)
            if len(args) == 2:
                x = resolve_length(args[0])
                y = resolve_length(args[1])
            elif len(args) == 1:
                x = resolve_length(args[0])
                y = 0
            else:
                raise ValueError(f"bad translation transform {tfstr}")

            transform = Transform.translation(x=x, y=y) @ transform

        elif tf_type == "translateX":  # SVG 2
            transform = Transform.translation(x=resolve_length(args), y=0) @ transform

        elif tf_type == "translateY":  # SVG 2
            transform = Transform.translation(x=0, y=resolve_length(args)) @ transform

    return transform


@force_nodocument
def svg_path_converter(pdf_path: PaintedPath, svg_path: str) -> None:
    pen = PathPen(pdf_path)
    parse_path(svg_path, pen)
    if not pen.first_is_move:
        raise ValueError("Path does not start with move item")


class SVGObject:
    """
    A representation of an SVG that has been converted to a PDF representation.
    """

    @classmethod
    def from_file(
        cls,
        filename: str | PathLike[str],
        *args: Any,
        encoding: str = "utf-8",
        **kwargs: Any,
    ) -> "SVGObject":
        """
        Create an `SVGObject` from the contents of the file at `filename`.

        Args:
            filename (path-like): the path to a file containing SVG data.
            *args: forwarded directly to the SVGObject initializer. For subclass use.
            encoding (str): optional charset encoding to use when reading the file.
            **kwargs: forwarded directly to the SVGObject initializer. For subclass use.

        Returns:
            A converted `SVGObject`.
        """
        with open(filename, "r", encoding=encoding) as svgfile:
            return cls(svgfile.read(), *args, **kwargs)

    def __init__(
        self, svg_text: str | bytes, image_cache: Optional[ImageCache] = None
    ) -> None:
        self.image_cache = image_cache  # Needed to render images
        self.cross_references: dict[str, Any] = {}
        self.css_class_styles: dict[str, dict[str, Any]] = {}
        self.gradient_definitions: dict[str, GradientPaint] = (
            {}
        )  # Store parsed gradients by ID
        self.width: Optional[Percent | float] = None
        self.height: Optional[Percent | float] = None
        # disabling bandit rule as we use defusedxml:
        svg_tree: "Element" = parse_xml_str(svg_text)  # nosec B314

        if svg_tree.tag not in xmlns_lookup("svg", "svg"):
            raise ValueError(f"root tag must be svg, not {svg_tree.tag}")

        self._collect_css_styles(svg_tree)
        self.extract_shape_info(svg_tree)
        self.convert_graphics(svg_tree)

    @force_nodocument
    def update_xref(self, key: Optional[str], referenced: Any) -> None:
        if key:
            key = "#" + key if not key.startswith("#") else key
            self.cross_references[key] = referenced

    def _collect_css_styles(self, root_tag: "Element") -> None:
        for node in root_tag.iter():
            if node.tag in xmlns_lookup("svg", "style"):
                css_text = "".join(node.itertext() or [])
                for class_name, declarations in _extract_css_class_styles(css_text):
                    existing = self.css_class_styles.setdefault(class_name, {})
                    existing.update(declarations)

    def _style_map_for(self, tag: "Element") -> dict[str, Any]:
        style_map: dict[str, Any] = {}
        if self.css_class_styles:
            class_attr = tag.attrib.get("class")
            if class_attr:
                for class_name in class_attr.split():
                    class_styles = self.css_class_styles.get(class_name)
                    if class_styles:
                        style_map.update(class_styles)
        inline = html.parse_css_style(tag.attrib.get("style", ""))
        for key, value in inline.items():
            norm_value = _normalize_css_value(value)
            if norm_value is not None:
                style_map[key] = norm_value
            elif key in style_map:
                style_map.pop(key, None)
        inheritable_attrs = (
            "font-family",
            "font-size",
            "font-style",
            "font-weight",
            "text-anchor",
            "white-space",
        )
        for attr in inheritable_attrs:
            if attr in tag.attrib:
                norm_value = _normalize_css_value(tag.attrib.get(attr))
                if norm_value is not None:
                    style_map.setdefault(attr, norm_value)
        return style_map

    @force_nodocument
    @staticmethod
    def _convert_gradient_coordinate(value: str, default: str = "0") -> float:
        """Convert SVG gradient coordinate (percentage or number) to float."""
        if value is None or value == "":
            value = default

        value = value.strip()

        if value.endswith("%"):
            return float(value[:-1]) / 100.0

        try:
            return float(value)
        except ValueError:
            try:
                return resolve_length(value)
            except ValueError:
                LOGGER.warning(
                    "Could not parse gradient coordinate '%s', using 0", value
                )
                return 0.0

    @force_nodocument
    @staticmethod
    def _parse_gradient_stops(
        gradient_element: "Element",
    ) -> list[tuple[float, DeviceRGB | DeviceGray]]:
        """Parse <stop> children of a gradient element."""
        stops: list[tuple[float, DeviceRGB | DeviceGray]] = []

        for stop_element in gradient_element:
            tag_name = without_ns(stop_element.tag)
            if tag_name != "stop":
                continue

            offset_str = stop_element.attrib.get("offset")
            if offset_str is None:
                LOGGER.warning("Found <stop> without offset, skipping")
                continue

            offset_str = offset_str.strip()
            if offset_str.endswith("%"):
                offset = float(offset_str[:-1]) / 100.0
            else:
                offset = float(offset_str)

            offset = max(0.0, min(1.0, offset))

            stop_color = None
            stop_opacity = 1.0

            style = stop_element.attrib.get("style", "")
            if style:
                style_dict = html.parse_css_style(style)
                stop_color = style_dict.get("stop-color")
                stop_opacity_str = style_dict.get("stop-opacity")
                if stop_opacity_str:
                    try:
                        stop_opacity = float(stop_opacity_str)
                    except ValueError:
                        LOGGER.warning(
                            "Invalid stop-opacity value: %s", stop_opacity_str
                        )

            if stop_color is None:
                stop_color = stop_element.attrib.get("stop-color", "black")

            if "stop-opacity" in stop_element.attrib:
                try:
                    stop_opacity = float(stop_element.attrib.get("stop-opacity") or "1")
                except ValueError:
                    pass

            try:
                color_obj = svgcolor(stop_color)

                if stop_opacity < 1.0:
                    if isinstance(color_obj, DeviceRGB):
                        color_obj = DeviceRGB(
                            color_obj.r, color_obj.g, color_obj.b, stop_opacity
                        )
                    if isinstance(color_obj, DeviceGray):
                        color_obj = DeviceGray(color_obj.g, stop_opacity)
                stops.append((offset, color_obj))

            except (ValueError, KeyError) as e:
                LOGGER.warning("Could not parse stop color '%s': %s", stop_color, e)
                continue

        return stops

    @force_nodocument
    @staticmethod
    def _extract_gradient_id(url_value: Optional[str]) -> Optional[str]:
        """Extract gradient ID from url(#id) format."""
        if not url_value or not isinstance(url_value, str):
            return None
        match = re.search(r'url\(\s*["\']?\s*#([^)"\'\s]+)', url_value)
        if match:
            return "#" + match.group(1)
        return None

    @force_nodocument
    def _parse_linear_gradient(self, grad_element: "Element") -> None:
        """Parse a <linearGradient> element and store it in gradient_definitions."""
        grad_id: Optional[str] = grad_element.attrib.get("id")
        if not grad_id:
            LOGGER.warning("Found <linearGradient> without id attribute, skipping")
            return

        if not grad_id.startswith("#"):
            grad_id = "#" + grad_id

        x1 = grad_element.attrib.get("x1", "0%")
        y1 = grad_element.attrib.get("y1", "0%")
        x2 = grad_element.attrib.get("x2", "100%")
        y2 = grad_element.attrib.get("y2", "0%")

        x1_val = self._convert_gradient_coordinate(x1, "0")
        y1_val = self._convert_gradient_coordinate(y1, "0")
        x2_val = self._convert_gradient_coordinate(x2, "1")
        y2_val = self._convert_gradient_coordinate(y2, "0")

        units_str = grad_element.attrib.get("gradientUnits", "objectBoundingBox")
        if units_str == "userSpaceOnUse":
            units = GradientUnits.USER_SPACE_ON_USE
        else:
            units = GradientUnits.OBJECT_BOUNDING_BOX

        spread_str = grad_element.attrib.get("spreadMethod", "pad")
        try:
            spread_method = GradientSpreadMethod.coerce(spread_str)
        except (ValueError, AttributeError):
            spread_method = GradientSpreadMethod.PAD
            LOGGER.warning("Invalid spreadMethod '%s', using PAD", spread_str)

        transform = None
        transform_str = grad_element.attrib.get("gradientTransform")
        if transform_str:
            try:
                transform = convert_transforms(transform_str)
            except (ValueError, AttributeError, TypeError) as e:
                LOGGER.warning("Could not parse gradientTransform: %s", e)

        stops = self._parse_gradient_stops(grad_element)

        if not stops:
            LOGGER.warning("Linear gradient '%s' has no valid stops, skipping", grad_id)
            return

        gradient = shape_linear_gradient(
            x1=x1_val,
            y1=y1_val,
            x2=x2_val,
            y2=y2_val,
            stops=stops,
            spread_method=spread_method,
        )

        gradient_paint = GradientPaint(
            gradient=gradient,
            units=units,
            gradient_transform=transform or Transform.identity(),
            spread_method=spread_method,
        )

        self.gradient_definitions[grad_id] = gradient_paint

        LOGGER.debug("Parsed linear gradient '%s' with %d stops", grad_id, len(stops))

    @force_nodocument
    def _parse_radial_gradient(self, grad_element: "Element") -> None:
        """Parse a <radialGradient> element and store it in gradient_definitions."""
        grad_id = grad_element.attrib.get("id")
        if not grad_id:
            LOGGER.warning("Found <radialGradient> without id attribute, skipping")
            return

        if not grad_id.startswith("#"):
            grad_id = "#" + grad_id

        cx = grad_element.attrib.get("cx", "50%")
        cy = grad_element.attrib.get("cy", "50%")
        r = grad_element.attrib.get("r", "50%")
        fx = grad_element.attrib.get("fx", cx)
        fy = grad_element.attrib.get("fy", cy)
        fr = grad_element.attrib.get("fr", "0%")

        cx_val = self._convert_gradient_coordinate(cx, "0.5")
        cy_val = self._convert_gradient_coordinate(cy, "0.5")
        r_val = self._convert_gradient_coordinate(r, "0.5")
        fx_val = self._convert_gradient_coordinate(fx, str(cx_val))
        fy_val = self._convert_gradient_coordinate(fy, str(cy_val))
        fr_val = self._convert_gradient_coordinate(fr, "0")

        if r_val <= 0:
            LOGGER.warning(
                "Radial gradient '%s' has invalid radius %s, skipping", grad_id, r_val
            )
            return

        units_str = grad_element.attrib.get("gradientUnits", "objectBoundingBox")
        units = (
            GradientUnits.USER_SPACE_ON_USE
            if units_str == "userSpaceOnUse"
            else GradientUnits.OBJECT_BOUNDING_BOX
        )

        spread_str = grad_element.attrib.get("spreadMethod", "pad")
        try:
            spread_method = GradientSpreadMethod.coerce(spread_str)
        except (ValueError, AttributeError):
            spread_method = GradientSpreadMethod.PAD

        transform = None
        transform_str = grad_element.attrib.get("gradientTransform")
        if transform_str:
            try:
                transform = convert_transforms(transform_str)
            except (ValueError, AttributeError, TypeError) as e:
                LOGGER.warning("Could not parse gradientTransform: %s", e)

        stops = self._parse_gradient_stops(grad_element)
        if not stops:
            LOGGER.warning("Radial gradient '%s' has no valid stops, skipping", grad_id)
            return

        gradient = shape_radial_gradient(
            cx=cx_val,
            cy=cy_val,
            r=r_val,
            stops=stops,
            fx=fx_val,
            fy=fy_val,
            fr=fr_val,
            spread_method=spread_method,
        )

        gradient_paint = GradientPaint(
            gradient=gradient,
            units=units,
            gradient_transform=transform or Transform.identity(),
            spread_method=spread_method,
        )

        self.gradient_definitions[grad_id] = gradient_paint
        LOGGER.debug("Parsed radial gradient '%s' with %d stops", grad_id, len(stops))

    @force_nodocument
    def _apply_gradient_paint(
        self,
        stylable: PaintedPath,
        svg_element: "Element",
        style_map: Optional[dict[str, Any]] = None,
    ) -> None:
        """Apply gradient paint to fill or stroke if a url(#gradientId) reference is found."""
        fill_value = _get_attr_or_style(svg_element, "fill", style_map)
        if fill_value:
            grad_id = self._extract_gradient_id(fill_value)
            if grad_id and grad_id in self.gradient_definitions:
                stylable.style.fill_color = self.gradient_definitions[grad_id]
                LOGGER.debug("Applied gradient %s to fill", grad_id)

        stroke_value = _get_attr_or_style(svg_element, "stroke", style_map)
        if stroke_value:
            grad_id = self._extract_gradient_id(stroke_value)
            if grad_id and grad_id in self.gradient_definitions:
                stylable.style.stroke_color = self.gradient_definitions[grad_id]
                LOGGER.debug("Applied gradient %s to stroke", grad_id)

    @force_nodocument
    def extract_shape_info(self, root_tag: "Element") -> None:
        """Collect shape info from the given SVG."""

        width_str = root_tag.get("width")
        height_str = root_tag.get("height")
        viewbox = root_tag.get("viewBox")
        # we don't fully support this, just check for its existence
        preserve_ar = root_tag.get("preserveAspectRatio", True)
        if preserve_ar == "none":
            self.preserve_ar = None
        else:
            self.preserve_ar = True

        self.width = None
        if width_str is not None:
            width_str.strip()
            if width_str.endswith("%"):
                self.width = Percent(width_str[:-1])
            else:
                self.width = resolve_length(width_str)

        self.height = None
        if height_str is not None:
            height_str.strip()
            if height_str.endswith("%"):
                self.height = Percent(height_str[:-1])
            else:
                self.height = resolve_length(height_str)

        if viewbox is None:
            self.viewbox = None
        else:
            viewbox = viewbox.strip()
            vx, vy, vw, vh = [float(num) for num in NUMBER_SPLIT.split(viewbox)]
            if (vw < 0) or (vh < 0):
                raise ValueError(f"invalid negative width/height in viewbox {viewbox}")

            self.viewbox = [vx, vy, vw, vh]

    @force_nodocument
    def convert_graphics(self, root_tag: "Element") -> None:
        """Convert the graphics contained in the SVG into the PDF representation."""
        base_group = GraphicsContext()
        base_group.style.stroke_width = None
        base_group.style.auto_close = False
        base_group.style.stroke_cap_style = StrokeCapStyle.BUTT

        self.build_group(root_tag, base_group)

        self.base_group = base_group

    def transform_to_page_viewport(
        self, pdf: "FPDF", align_viewbox: bool = True
    ) -> tuple[float, float, GraphicsContext]:
        """
        Size the converted SVG paths to the page viewport.

        The SVG document size can be specified relative to the rendering viewport
        (e.g. width=50%). If the converted SVG sizes are relative units, then this
        computes the appropriate scale transform to size the SVG to the correct
        dimensions for a page in the current PDF document.

        If the SVG document size is specified in absolute units, then it is not scaled.

        Args:
            pdf (fpdf.fpdf.FPDF): the pdf to use the page size of.
            align_viewbox (bool): if True, mimic some of the SVG alignment rules if the
                viewbox aspect ratio does not match that of the viewport.

        Returns:
            The same thing as `SVGObject.transform_to_rect_viewport`.
        """

        return self.transform_to_rect_viewport(pdf.k, pdf.epw, pdf.eph, align_viewbox)

    def transform_to_rect_viewport(
        self,
        scale: float,
        width: float | Percent,
        height: float | Percent,
        align_viewbox: bool = True,
        ignore_svg_top_attrs: bool = False,
    ) -> tuple[float, float, GraphicsContext]:
        """
        Size the converted SVG paths to an arbitrarily sized viewport.

        The SVG document size can be specified relative to the rendering viewport
        (e.g. width=50%). If the converted SVG sizes are relative units, then this
        computes the appropriate scale transform to size the SVG to the correct
        dimensions for a page in the current PDF document.

        Args:
            scale (Number): the scale factor from document units to PDF points.
            width (Number): the width of the viewport to scale to in document units.
            height (Number): the height of the viewport to scale to in document units.
            align_viewbox (bool): if True, mimic some of the SVG alignment rules if the
                viewbox aspect ratio does not match that of the viewport.
            ignore_svg_top_attrs (bool): ignore <svg> top attributes like "width", "height"
                or "preserveAspectRatio" when figuring the image dimensions.
                Require width & height to be provided as parameters.

        Returns:
            A tuple of (width, height, `fpdf.drawing.GraphicsContext`), where width and
            height are the resolved width and height (they may be 0. If 0, the returned
            `fpdf.drawing.GraphicsContext` will be empty). The
            `fpdf.drawing.GraphicsContext` contains all of the paths that were
            converted from the SVG, scaled to the given viewport size.
        """
        vp_width: float
        vp_height: float

        if ignore_svg_top_attrs:
            # width/height may be Percent when passed from FPDF.image; resolve to float
            if isinstance(width, Percent):
                vp_width = float(width)
            else:
                vp_width = float(width)
        elif isinstance(self.width, Percent):
            if not width:
                raise ValueError(
                    'SVG "width" is a percentage, hence a viewport width is required'
                )
            vp_width = self.width * width / 100
        else:
            vp_width = self.width or width

        if ignore_svg_top_attrs:
            if isinstance(height, Percent):
                vp_height = float(height)
            else:
                vp_height = float(height)
        elif isinstance(self.height, Percent):
            if not height:
                raise ValueError(
                    'SVG "height" is a percentage, hence a viewport height is required'
                )
            vp_height = self.height * height / 100
        else:
            vp_height = self.height or height

        if scale == 1:
            transform = Transform.identity()
        else:
            transform = Transform.scaling(1 / scale)

        if self.viewbox:
            vx, vy, vw, vh = self.viewbox

            if (vw == 0) or (vh == 0):
                return 0, 0, GraphicsContext()

            w_ratio = vp_width / vw
            h_ratio = vp_height / vh

            if not ignore_svg_top_attrs and self.preserve_ar and (w_ratio != h_ratio):
                w_ratio = h_ratio = min(w_ratio, h_ratio)

            transform = (
                transform
                @ Transform.translation(x=-vx, y=-vy)
                @ Transform.scaling(x=w_ratio, y=h_ratio)
            )

            if align_viewbox:
                transform = transform @ Transform.translation(
                    x=vp_width / 2 - (vw / 2) * w_ratio,
                    y=vp_height / 2 - (vh / 2) * h_ratio,
                )

        self.base_group.transform = transform

        return vp_width / scale, vp_height / scale, self.base_group

    def draw_to_page(
        self,
        pdf: "FPDF",
        x: Optional[float] = None,
        y: Optional[float] = None,
        debug_stream: Optional[bool] = None,
    ) -> None:
        """
        Directly draw the converted SVG to the given PDF's current page.

        The page viewport is used for sizing the SVG.

        Args:
            pdf (fpdf.fpdf.FPDF): the document to which the converted SVG is rendered.
            x (Number): abscissa of the converted SVG's top-left corner.
            y (Number): ordinate of the converted SVG's top-left corner.
            debug_stream (io.TextIO): *DEPRECATED* the stream to which rendering debug info will be
                written.
        """
        self.image_cache = pdf.image_cache  # Needed to render images
        _, _, path = self.transform_to_page_viewport(pdf)

        old_x, old_y = pdf.x, pdf.y
        try:
            if x is not None and y is not None:
                pdf.set_xy(0, 0)
                assert path.transform is not None
                path.transform = path.transform @ Transform.translation(x, y)

            pdf.draw_path(path, debug_stream)

        finally:
            pdf.set_xy(old_x, old_y)

    # defs paths are not drawn immediately but are added to xrefs and can be referenced
    # later to be drawn.
    @force_nodocument
    def handle_defs(self, defs: "Element") -> None:
        """Produce lookups for groups and paths inside the <defs> tag"""
        for child in defs:
            if child.tag in xmlns_lookup("svg", "g"):
                self.build_group(child)
            elif child.tag in xmlns_lookup("svg", "a"):
                # <a> tags aren't supported but we need to recurse into them to
                # render nested elements.
                LOGGER.warning(
                    "Ignoring unsupported SVG tag: <a> (contributions are welcome to add support for it)",
                )
                self.build_group(child)
            elif child.tag in xmlns_lookup("svg", "path"):
                self.build_path(child)
            elif child.tag in xmlns_lookup("svg", "image"):
                self.build_image(child)
            elif child.tag in shape_tags:
                self.build_shape(child)
            elif child.tag in xmlns_lookup("svg", "linearGradient"):
                self._parse_linear_gradient(child)
            elif child.tag in xmlns_lookup("svg", "radialGradient"):
                self._parse_radial_gradient(child)
            elif child.tag in xmlns_lookup("svg", "clipPath"):
                try:
                    clip_id = child.attrib["id"]
                except KeyError:
                    clip_id = None
                for child_ in child:
                    self.build_clipping_path(child_, clip_id)
            elif child.tag in xmlns_lookup("svg", "style"):
                # Styles handled globally during parsing
                continue
            else:
                LOGGER.warning(
                    "Ignoring unsupported SVG tag: <%s> (contributions are welcome to add support for it)",
                    without_ns(child.tag),
                )

    # this assumes xrefs only reference already-defined ids.
    # I don't know if this is required by the SVG spec.
    @force_nodocument
    def build_xref(self, xref: "Element") -> GraphicsContext:
        """Resolve a cross-reference to an already-seen SVG element by ID."""
        style_map = self._style_map_for(xref)
        pdf_group = GraphicsContext()
        apply_styles(pdf_group, xref, style_map)

        for candidate in xmlns_lookup("xlink", "href", "id"):
            try:
                ref = xref.attrib[candidate]
                break
            except KeyError:
                pass
        else:
            raise ValueError(f"use {xref} doesn't contain known xref attribute")

        try:
            pdf_group.add_item(self.cross_references[ref])
        except KeyError:
            raise ValueError(
                f"use {xref} references nonexistent ref id {ref}"
            ) from None

        if "x" in xref.attrib or "y" in xref.attrib:
            # Quoting the SVG spec - 5.6.2. Layout of re-used graphics:
            # > The x and y properties define an additional transformation translate(x,y)
            x, y = float(xref.attrib.get("x", 0)), float(xref.attrib.get("y", 0))
            pdf_group.transform = Transform.translation(x=x, y=y)
        # Note that we currently do not support "width" & "height" in <use>

        return pdf_group

    @force_nodocument
    def build_group(
        self,
        group: "Element",
        pdf_group: Optional[GraphicsContext] = None,
        inherited_style: Optional[dict[str, Any]] = None,
    ) -> GraphicsContext:
        """Handle nested items within a group <g> tag."""
        local_style = self._style_map_for(group)
        merged_style = dict(inherited_style or {})
        merged_style.update(local_style)
        if pdf_group is None:
            pdf_group = GraphicsContext()
        apply_styles(pdf_group, group, merged_style)

        # handle defs before anything else
        for child in [
            child for child in group if child.tag in xmlns_lookup("svg", "defs")
        ]:
            self.handle_defs(child)

        for child in group:
            if child.tag in xmlns_lookup("svg", "defs"):
                self.handle_defs(child)
            elif child.tag in xmlns_lookup("svg", "style"):
                # Stylesheets already parsed globally.
                continue
            elif child.tag in xmlns_lookup("svg", "g"):
                pdf_group.add_item(self.build_group(child, None, merged_style), False)
            elif child.tag in xmlns_lookup("svg", "a"):
                # <a> tags aren't supported but we need to recurse into them to
                # render nested elements.
                LOGGER.warning(
                    "Ignoring unsupported SVG tag: <a> (contributions are welcome to add support for it)",
                )
                pdf_group.add_item(self.build_group(child, None, merged_style), False)
            elif child.tag in xmlns_lookup("svg", "path"):
                pdf_group.add_item(self.build_path(child), False)
            elif child.tag in shape_tags:
                pdf_group.add_item(self.build_shape(child), False)
            elif child.tag in xmlns_lookup("svg", "use"):
                pdf_group.add_item(self.build_xref(child), False)
            elif child.tag in xmlns_lookup("svg", "image"):
                pdf_group.add_item(self.build_image(child), False)
            elif child.tag in xmlns_lookup("svg", "text"):
                text_path = self.build_text(child, merged_style)
                if text_path:
                    pdf_group.add_item(text_path, False)
            else:
                LOGGER.warning(
                    "Ignoring unsupported SVG tag: <%s> (contributions are welcome to add support for it)",
                    without_ns(child.tag),
                )

        self.update_xref(group.attrib.get("id"), pdf_group)

        return pdf_group

    @force_nodocument
    def build_path(self, path: "Element") -> PaintedPath:
        """Convert an SVG <path> tag into a PDF path object."""
        style_map = self._style_map_for(path)
        pdf_path = PaintedPath()
        apply_styles(pdf_path, path, style_map)
        self._apply_gradient_paint(pdf_path, path, style_map)
        self.apply_clipping_path(pdf_path, path, style_map)
        svg_path = path.attrib.get("d")
        if svg_path is not None:
            svg_path_converter(pdf_path, svg_path)
        self.update_xref(path.attrib.get("id"), pdf_path)
        return pdf_path

    @force_nodocument
    def build_shape(self, shape: "Element") -> PaintedPath:
        """Convert an SVG shape tag into a PDF path object. Necessary to make xref (because ShapeBuilder doesn't have access to this object.)"""
        style_map = self._style_map_for(shape)
        shape_builder = getattr(ShapeBuilder, shape_tags[shape.tag])
        shape_path = shape_builder(shape)
        assert isinstance(shape_path, PaintedPath)
        apply_styles(shape_path, shape, style_map)
        self._apply_gradient_paint(shape_path, shape, style_map)
        self.apply_clipping_path(shape_path, shape, style_map)
        self.update_xref(shape.attrib.get("id"), shape_path)
        return shape_path

    @force_nodocument
    def build_clipping_path(self, shape: "Element", clip_id: Optional[str]) -> None:
        if shape.tag in shape_tags:
            style_map = self._style_map_for(shape)
            shape_builder = getattr(ShapeBuilder, shape_tags[shape.tag])
            clipping_path_shape = shape_builder(shape, True)
            apply_styles(clipping_path_shape, shape, style_map)
        elif shape.tag in xmlns_lookup("svg", "path"):
            style_map = self._style_map_for(shape)
            clipping_path_shape = PaintedPath()
            apply_styles(clipping_path_shape, shape, style_map)
            clipping_path_shape.paint_rule = PathPaintRule.DONT_PAINT
            svg_path = shape.attrib.get("d")
            if svg_path is not None:
                svg_path_converter(clipping_path_shape, svg_path)
        else:
            LOGGER.warning(
                "Ignoring unsupported <clipPath> child tag: <%s> (contributions are welcome to add support for it)",
                without_ns(shape.tag),
            )
            return
        self.update_xref(clip_id, clipping_path_shape)

    @force_nodocument
    def apply_clipping_path(
        self,
        stylable: PaintedPath,
        svg_element: "Element",
        style_map: Optional[dict[str, Any]] = None,
    ) -> None:
        clip_value = None
        if style_map and "clip-path" in style_map:
            clip_value = style_map["clip-path"]
        if clip_value is None:
            clip_value = svg_element.attrib.get("clip-path")
        if clip_value:
            clipping_path_id = re.search(r"url\((\#\w+)\)", clip_value)
            assert clipping_path_id is not None
            stylable.clipping_path = self.cross_references[clipping_path_id[1]]

    @force_nodocument
    def build_image(self, image: "Element") -> "SVGImage":
        href = None
        for key in xmlns_lookup("xlink", "href"):
            if key in image.attrib:
                href = image.attrib[key]
                break
        if not href:
            raise ValueError("<image> is missing a href attribute")
        width = float(image.attrib.get("width", 0))
        height = float(image.attrib.get("height", 0))
        if "preserveAspectRatio" in image.attrib:
            LOGGER.warning(
                '"preserveAspectRatio" defined on <image> is currently not supported (contributions are welcome to add support for it)'
            )
        if "style" in image.attrib:
            LOGGER.warning(
                '"style" defined on <image> is currently not supported (contributions are welcome to add support for it)'
            )
        if "transform" in image.attrib:
            LOGGER.warning(
                '"transform" defined on <image> is currently not supported (contributions are welcome to add support for it)'
            )
        # Note: at this moment, self.image_cache is not set yet:
        svg_image = SVGImage(
            href=href,
            x=float(image.attrib.get("x", "0")),
            y=float(image.attrib.get("y", "0")),
            width=width,
            height=height,
            svg_obj=self,
        )
        self.update_xref(image.attrib.get("id"), svg_image)
        return svg_image

    @force_nodocument
    def build_text(
        self, text_tag: "Element", inherited_style: Optional[dict[str, Any]] = None
    ) -> Optional[PaintedPath]:
        """
        Convert <text> (and simple <tspan>) into a PaintedPath with Text runs.
        - Uses Text baseline at (x,y)
        - Honors x/y and dx/dy on <text> and direct child <tspan>
        - Flattens nested tspans; advanced per-character positioning is not implemented
        """
        local_style = self._style_map_for(text_tag)
        effective_style = dict(inherited_style or {})
        effective_style.update(local_style)
        path = PaintedPath()
        apply_styles(path, text_tag, local_style)
        self.apply_clipping_path(path, text_tag, effective_style)

        preserve_parent = _preserve_ws(effective_style, text_tag)

        base_family, base_emph, base_size, base_anchor = _parse_font_attrs(
            text_tag, effective_style
        )
        if base_family is None:
            base_family = "sans-serif"
        default_font_size = (
            base_size if base_size is not None else resolve_length("16px")
        )
        base_x, base_y, base_dx, base_dy = _parse_xy_delta(
            text_tag, effective_style, font_size=default_font_size
        )
        anchor_x = base_x + base_dx
        anchor_y = base_y + base_dy

        text_runs: list[TextRun] = []
        pending_dx = 0.0
        pending_dy = 0.0

        def _style_for_run(
            tag: Optional["Element"], style_map_for_tag: Optional[dict[str, Any]]
        ) -> Optional[GraphicsStyle]:
            if tag is None or style_map_for_tag is None:
                return None
            context = GraphicsContext()
            apply_styles(context, tag, style_map_for_tag)
            context.style.auto_close = GraphicsStyle.INHERIT
            overrides = any(
                getattr(context.style, prop) is not GraphicsStyle.INHERIT
                for prop in GraphicsStyle.MERGE_PROPERTIES
            )
            if not overrides:
                return None
            return deepcopy(context.style)

        def _add_run(
            raw_text: str,
            family: Optional[str],
            emphasis: Optional[str],
            size: Optional[float],
            preserve: bool,
            dx_extra: float = 0.0,
            dy_extra: float = 0.0,
            abs_x: Optional[float] = None,
            abs_y: Optional[float] = None,
            style_tag: Optional["Element"] = None,
            style_map_for_tag: Optional[dict[str, Any]] = None,
        ) -> None:
            nonlocal pending_dx, pending_dy
            raw = raw_text or ""
            collapsed = _collapse_ws(raw, preserve=preserve)
            if preserve:
                content = collapsed
            else:
                trimmed = collapsed.strip()
                if trimmed:
                    content = trimmed
                    if raw[:1].isspace():
                        content = " " + content
                    if raw[-1:].isspace():
                        content = content + " "
                else:
                    if (raw[:1].isspace() or raw[-1:].isspace()) and text_runs:
                        content = " "
                    else:
                        content = ""
            if not content:
                pending_dx += dx_extra
                pending_dy += dy_extra
                return

            run_size = size if size is not None else default_font_size
            run_family = family or base_family or "sans-serif"
            run_emphasis = (emphasis if emphasis is not None else base_emph) or ""
            run_dx = pending_dx + dx_extra
            run_dy = pending_dy + dy_extra
            pending_dx = 0.0
            pending_dy = 0.0
            run_style = _style_for_run(style_tag, style_map_for_tag)

            text_runs.append(
                TextRun(
                    text=content,
                    family=run_family,
                    emphasis=run_emphasis,
                    size=run_size,
                    dx=run_dx,
                    dy=run_dy,
                    abs_x=abs_x,
                    abs_y=abs_y,
                    run_style=run_style,
                )
            )

        # Leading text (before child <tspan>)
        _add_run(
            text_tag.text or "",
            base_family,
            base_emph,
            base_size,
            preserve=preserve_parent,
            style_tag=None,
            style_map_for_tag=None,
        )

        for child in text_tag:
            if child.tag in xmlns_lookup("svg", "tspan"):
                child_local_style = self._style_map_for(child)
                child_effective_style = dict(effective_style)
                child_effective_style.update(child_local_style)
                fam, emph, size, _anchor = _parse_font_attrs(
                    child, child_effective_style
                )
                run_font_size = size if size is not None else default_font_size
                x, y, dx, dy = _parse_xy_delta(
                    child, child_effective_style, font_size=run_font_size
                )

                child_preserve = _preserve_ws(child_effective_style, child)
                raw_itertext = "".join(child.itertext())
                tail_text = child.tail or ""
                if tail_text and raw_itertext.endswith(tail_text):
                    run_text = raw_itertext[: -len(tail_text)]
                else:
                    run_text = raw_itertext
                abs_x = None
                abs_y = None
                if "x" in child.attrib or (
                    child_local_style is not None and "x" in child_local_style
                ):
                    abs_x = x
                if "y" in child.attrib or (
                    child_local_style is not None and "y" in child_local_style
                ):
                    abs_y = y

                _add_run(
                    run_text,
                    fam,
                    emph,
                    size,
                    preserve=child_preserve,
                    dx_extra=dx,
                    dy_extra=dy,
                    abs_x=abs_x,
                    abs_y=abs_y,
                    style_tag=child,
                    style_map_for_tag=child_local_style,
                )

                # Text between tspans inherits parent style
                _add_run(
                    child.tail or "",
                    base_family,
                    base_emph,
                    base_size,
                    preserve=preserve_parent,
                    style_tag=None,
                    style_map_for_tag=None,
                )
            else:
                # other child tags are ignored (already logged elsewhere)
                pass

        if text_runs:
            path.add_path_element(
                Text(
                    x=anchor_x,
                    y=anchor_y,
                    text_runs=tuple(text_runs),
                    text_anchor=base_anchor,
                ),
                clone=False,
            )

        self.update_xref(text_tag.attrib.get("id"), path)
        return path


class SVGImage(NamedTuple):
    href: str
    x: float
    y: float
    width: float
    height: float
    svg_obj: SVGObject

    def __deepcopy__(self: "SVGImage", _memo: dict[int, Any]) -> "SVGImage":
        # Defining this method is required to avoid the .svg_obj reference to be cloned:
        return SVGImage(
            href=self.href,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            svg_obj=self.svg_obj,
        )

    # pylint: disable=unused-argument
    @force_nodocument
    def render(
        self,
        resource_registry: "ResourceCatalog",
        style: GraphicsStyle,
        last_item: "Renderable",
        initial_point: Point,
    ) -> tuple[str, "Renderable", Point]:
        image_cache = self.svg_obj and self.svg_obj.image_cache
        if not image_cache:
            raise AssertionError(
                "fpdf2 bug - Cannot render a raster image without a SVGObject.image_cache"
            )

        # We lazy-import this function to circumvent a circular import problem:
        # pylint: disable=cyclic-import,import-outside-toplevel
        from .image_parsing import preload_image

        _, _, info = preload_image(image_cache, self.href)
        if isinstance(info, VectorImageInfo):
            LOGGER.warning(
                "Inserting .svg vector graphics in <image> tags is currently not supported (contributions are welcome to add support for it)"
            )
            return "", last_item, initial_point
        w, h = info.size_in_document_units(self.width, self.height)
        stream_content = stream_content_for_raster_image(
            info=info,
            x=self.x,
            y=self.y,
            w=w,
            h=h,
            keep_aspect_ratio=True,
        )
        return stream_content, last_item, initial_point

    @property
    def end_point(self) -> Point:
        return Point(self.x, self.y)

    # pylint: disable=unused-argument
    def bounding_box(self, start: Point) -> tuple[BoundingBox, Point]:
        x0 = self.x
        y0 = self.y
        x1 = self.x + self.width
        y1 = self.y + self.height

        bbox = BoundingBox.from_points(
            [
                Point(x0, y0),
                Point(x1, y0),
                Point(x0, y1),
                Point(x1, y1),
            ]
        )
        return bbox, Point(self.x, self.y)



import math
from pathlib import Path

import pytest
from fpdf import FPDF
from fpdf.pattern import (
    LinearGradient,
    RadialGradient,
    SweepGradient,
    shape_linear_gradient,
    shape_radial_gradient,
)
from fpdf.drawing_primitives import DeviceRGB
from fpdf.drawing import PaintedPath, Transform, GradientPaint
from fpdf.enums import GradientSpreadMethod

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def _new_pdf():
    pdf = FPDF(unit="pt", format="A4")
    pdf.add_page()
    return pdf


def _rect(x, y, w, h):
    return PaintedPath().rectangle(x, y, w, h)


def _circle(cx, cy, r):
    circle = PaintedPath().move_to(cx + r, cy)
    circle.circle(cx, cy, r)
    return circle


def test_linear_gradient_fill_rotated_vs_user_space(tmp_path):
    pdf = _new_pdf()

    # Path A with objectBoundingBox + rotation
    path = _rect(10, 20, 100, 50)
    gradient = LinearGradient(
        0, 0, 1, 0, colors=["#FF0000", "#0000FF"], extend_before=True, extend_after=True
    )
    matrix_rotation = Transform.rotation(math.radians(30))
    paint = GradientPaint(
        gradient, units="objectBoundingBox", gradient_transform=matrix_rotation
    )
    path.style.fill_color = paint
    path.style.stroke_color = None

    # Path B with userSpaceOnUse (absolute coords) - rendered as DeviceGray
    path2 = _rect(10, 90, 100, 50)
    gradient2 = LinearGradient(
        10,
        0,
        110,
        0,
        colors=["#ffffff", "#000000"],
        extend_before=True,
        extend_after=True,
    )
    paint2 = GradientPaint(gradient2, units="userSpaceOnUse")
    path2.style.fill_color = paint2
    path2.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(path)
        dc.add_item(path2)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_linear_rotated_vs_user_space.pdf",
        tmp_path,
    )


def test_linear_gradient_objbox_scale_translate(tmp_path: Path):
    pdf = _new_pdf()

    path = _rect(150, 40, 180, 90)  # non-uniform aspect
    lg = LinearGradient(
        0, 0, 1, 1, colors=["#0000FF", "#FFFFFF", "#FF0000"], extend_after=True
    )
    M = Transform.scaling(1.2, 0.6) @ Transform.translation(
        0.15, -0.1
    )  # in gradient space
    path.style.fill_color = GradientPaint(
        lg, units="objectBoundingBox", gradient_transform=M
    )
    path.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(path)

    assert_pdf_equal(
        pdf, HERE / "generated_pdf" / "gradient_linear_scale_translate.pdf", tmp_path
    )


def test_linear_gradient_userspace_custom_pivot(tmp_path: Path):
    pdf = _new_pdf()

    path = _rect(40, 160, 220, 60)
    gradient = LinearGradient(40, 160, 260, 220, colors=["#222222", "#DDDDDD"])
    path.style.fill_color = GradientPaint(
        gradient,
        units="userSpaceOnUse",
        gradient_transform=Transform.translation(-150, -190)
        .rotate(math.radians(-25))
        .translate(150, 190),
    )
    path.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(path)

    assert_pdf_equal(
        pdf, HERE / "generated_pdf" / "gradient_userspace_pivot.pdf", tmp_path
    )


@pytest.mark.parametrize(
    "extend_before,extend_after,basename",
    [
        (False, False, "stops_no_extend"),
        (True, False, "stops_extend_before"),
        (False, True, "stops_extend_after"),
        (True, True, "stops_extend_both"),
    ],
)
def test_linear_gradient_color_stops_extends(
    extend_before, extend_after, basename, tmp_path
):
    pdf = _new_pdf()

    path = _rect(30, 30, 260, 40)
    path.style.stroke_color = "#000000"
    path.style.stroke_width = 2
    gradient = LinearGradient(
        50,
        0,
        270,
        0,
        colors=["#0000FF", "#00FFFF", "#00FF00", "#FFFF00", "#FF0000"],
        bounds=[0.25, 0.50, 0.75],
        extend_before=extend_before,
        extend_after=extend_after,
    )
    paint = GradientPaint(gradient, units="userSpaceOnUse")
    path.style.fill_color = paint

    with pdf.drawing_context() as dc:
        dc.add_item(path)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / f"gradient_linear_{basename}.pdf",
        tmp_path,
    )


def test_radial_gradient_objbox_vs_userspace(tmp_path: Path):
    pdf = _new_pdf()

    path1 = _circle(140, 100, 50)
    gradient1 = RadialGradient(
        0.5, 0.5, 0.0, 0.5, 0.5, 0.5, colors=["#FFFFFF", "#0000AA"]
    )
    path1.style.fill_color = GradientPaint(gradient1, units="objectBoundingBox")
    path1.style.stroke_color = None

    path2 = _circle(340, 100, 50)
    gradient2 = RadialGradient(340, 100, 0, 340, 100, 60, colors=["#FFFFFF", "#AA0000"])
    path2.style.fill_color = GradientPaint(gradient2, units="userSpaceOnUse")
    path2.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(path1)
        dc.add_item(path2)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_radial_objbox_userspace.pdf",
        tmp_path,
    )


def test_radial_gradient_focal_offset_extend(tmp_path: Path):
    pdf = _new_pdf()

    path = _rect(40, 100, 260, 120)
    # Circle center at (0.6, 0.5), focal at (0.3, 0.45), radius 0.7 (object bbox units)
    gradient = RadialGradient(
        0.6,
        0.5,
        0.7,
        0.3,
        0.45,
        0.0,
        colors=["#FFFFFF", "#00AA00", "#003300"],
        bounds=[0.6],
        extend_before=True,
        extend_after=True,
    )
    path.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")
    path.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(path)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_radial_focal_extend.pdf",
        tmp_path,
    )


def test_shared_gradient_instance(tmp_path: Path):
    pdf = _new_pdf()

    gradient = LinearGradient(0, 0, 1, 0, colors=["#FF0000", "#0000FF"])

    path1 = _rect(40, 40, 100, 40)
    path1.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")

    path2 = _rect(160, 40, 100, 40)

    # Different gradient transform but same LinearGradient object
    path2.style.fill_color = GradientPaint(
        gradient,
        units="objectBoundingBox",
        gradient_transform=Transform.translation(-0.5, -0.5)
        .rotate(math.radians(180))
        .translate(0.5, 0.5),
    )

    with pdf.drawing_context() as dc:
        dc.add_item(path1)
        dc.add_item(path2)

    assert_pdf_equal(pdf, HERE / "generated_pdf" / "gradient_shared.pdf", tmp_path)


def test_gradient_fill_with_solid_stroke(tmp_path: Path):
    pdf = _new_pdf()

    gradient = LinearGradient(0, 0, 1, 0, colors=["#FFFFFF", "#000000"])

    path1 = _rect(50, 20, 120, 50)
    path1.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")
    path1.style.stroke_color = "#0000FF"
    path1.style.stroke_width = 3.0

    path2 = _rect(200, 20, 120, 50)
    path2.style.fill_color = "#0000FF"
    path2.style.stroke_color = GradientPaint(gradient, units="objectBoundingBox")
    path2.style.stroke_width = 3.0

    path3 = _rect(350, 20, 120, 50)
    path3.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")
    path3.style.stroke_color = GradientPaint(gradient, units="objectBoundingBox")
    path3.style.stroke_width = 3.0

    with pdf.drawing_context() as dc:
        dc.add_item(path1)
        dc.add_item(path2)
        dc.add_item(path3)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_with_stroke.pdf",
        tmp_path,
    )


def test_gradients_across_pages_resource_reuse(tmp_path: Path):
    pdf = FPDF(unit="pt", format="A4")
    pdf.set_compression(False)

    # Page 1
    pdf.add_page()
    path1 = _rect(60, 100, 220, 70)
    gradient = LinearGradient(0, 0, 1, 0, colors=["#222222", "#EEEEEE"])
    path1.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")
    path1.style.stroke_color = None
    with pdf.drawing_context() as dc:
        dc.add_item(path1)

    # Page 2 (same gradient)
    pdf.add_page()
    path2 = _rect(60, 100, 220, 70)
    path2.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")
    path2.style.stroke_color = None
    with pdf.drawing_context() as dc:
        dc.add_item(path2)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_linear_different_pages.pdf",
        tmp_path,
    )


def test_gradient_shape_linear(tmp_path: Path):
    pdf = _new_pdf()

    # SVG-like stops
    stops = [
        (0.00, "#000000"),
        (0.50, "#ff0000"),
        (1.00, "#00ff00"),
    ]

    gradient = shape_linear_gradient(0, 0, 1, 0, stops)

    r1 = _rect(40, 100, 120, 60)
    r1.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")
    r1.style.stroke_color = "#000000"
    r1.style.stroke_width = 1.0

    r2 = _rect(180, 100, 120, 60)
    r2.style.fill_color = GradientPaint(
        gradient,
        units="objectBoundingBox",
        gradient_transform=Transform.translation(-0.5, -0.5)
        .rotate(math.radians(180))
        .translate(0.5, 0.5),
    )
    r2.style.stroke_color = "#000000"
    r2.style.stroke_width = 1.0

    r3 = _rect(320, 100, 120, 60)
    r3.style.fill_color = GradientPaint(
        gradient,
        units="objectBoundingBox",
        gradient_transform=Transform.scaling(0.4, 1.0).translate(0.3, 0.0),
    )
    r3.style.stroke_color = "#000000"
    r3.style.stroke_width = 1.0

    with pdf.drawing_context() as dc:
        dc.add_item(r1)
        dc.add_item(r2)
        dc.add_item(r3)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_shape_linear.pdf",
        tmp_path,
    )


def test_gradient_shape_radial(tmp_path: Path):
    pdf = _new_pdf()

    # Stops from white center to blue edge
    stops = [
        (0.00, "#ffffff"),
        (0.75, "#66aaff"),
        (1.00, "#002266"),
    ]

    rg_center = shape_radial_gradient(cx=0.5, cy=0.5, r=0.5, stops=stops)

    c1 = _circle(110, 150, 50)
    c1.style.fill_color = GradientPaint(rg_center, units="objectBoundingBox")
    c1.style.stroke_color = None

    rg_focal = shape_radial_gradient(
        cx=0.5, cy=0.5, r=0.5, stops=stops, fx=0.3, fy=0.4, fr=0.0
    )

    c2 = _circle(260, 150, 50)
    c2.style.fill_color = GradientPaint(rg_focal, units="objectBoundingBox")
    c2.style.stroke_color = None

    rg_rect = shape_radial_gradient(
        cx=0.5, cy=0.5, r=0.7, stops=stops, fx=0.35, fy=0.45, fr=0.0
    )

    r = _rect(320, 100, 120, 120)
    r.style.fill_color = GradientPaint(rg_rect, units="objectBoundingBox")
    r.style.stroke_color = "#000000"
    r.style.stroke_width = 0.8

    with pdf.drawing_context() as dc:
        dc.add_item(c1)
        dc.add_item(c2)
        dc.add_item(r)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_shape_radial.pdf",
        tmp_path,
    )


def test_linear_gradient_spread_methods(tmp_path: Path):
    pdf = _new_pdf()

    def build_paint(spread_method: GradientSpreadMethod) -> GradientPaint:
        return GradientPaint(
            LinearGradient(
                0,
                0,
                1,
                0,
                colors=["#ff6600", "#003366"],
                extend_before=False,
                extend_after=False,
            ),
            units="objectBoundingBox",
            gradient_transform=Transform.scaling(0.4, 1.0).translate(0.3, 0.0),
            spread_method=spread_method,
        )

    path_pad = _rect(40, 60, 120, 80)
    path_pad.style.fill_color = build_paint(GradientSpreadMethod.PAD)
    path_pad.style.stroke_color = "#000000"
    path_pad.style.stroke_width = 0.8

    path_repeat = _rect(200, 60, 120, 80)
    path_repeat.style.fill_color = build_paint(GradientSpreadMethod.REPEAT)
    path_repeat.style.stroke_color = "#000000"
    path_repeat.style.stroke_width = 0.8

    path_reflect = _rect(360, 60, 120, 80)
    path_reflect.style.fill_color = build_paint(GradientSpreadMethod.REFLECT)
    path_reflect.style.stroke_color = "#000000"
    path_reflect.style.stroke_width = 0.8

    with pdf.drawing_context() as dc:
        dc.add_item(path_pad)
        dc.add_item(path_repeat)
        dc.add_item(path_reflect)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_linear_spread_methods.pdf",
        tmp_path,
    )


def test_radial_gradient_spread_methods(tmp_path: Path):
    pdf = _new_pdf()

    def build_paint(spread_method: GradientSpreadMethod) -> GradientPaint:
        return GradientPaint(
            RadialGradient(
                0.5,
                0.5,
                0.0,
                0.5,
                0.5,
                0.3,
                colors=["#ffffcc", "#ff6600", "#660000"],
                extend_before=False,
                extend_after=False,
            ),
            units="objectBoundingBox",
            spread_method=spread_method,
        )

    circle_pad = _circle(110, 200, 70)
    circle_pad.style.fill_color = build_paint(GradientSpreadMethod.PAD)
    circle_pad.style.stroke_color = None

    circle_repeat = _circle(260, 200, 70)
    circle_repeat.style.fill_color = build_paint(GradientSpreadMethod.REPEAT)
    circle_repeat.style.stroke_color = None

    circle_reflect = _circle(410, 200, 70)
    circle_reflect.style.fill_color = build_paint(GradientSpreadMethod.REFLECT)
    circle_reflect.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(circle_pad)
        dc.add_item(circle_repeat)
        dc.add_item(circle_reflect)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_radial_spread_methods.pdf",
        tmp_path,
    )


def test_sweep_gradient_types(tmp_path: Path):
    pdf = _new_pdf()

    stops = [
        (0.0, "#ff0000"),
        (0.5, "#00ff00"),
        (1.0, "#0000ff"),
    ]

    circle_pad = _circle(120, 160, 60)
    circle_pad.style.fill_color = GradientPaint(
        SweepGradient(
            cx=0.5,
            cy=0.5,
            start_angle=0.0,
            end_angle=math.tau,
            stops=stops,
        ),
        units="objectBoundingBox",
        spread_method=GradientSpreadMethod.PAD,
    )
    circle_pad.style.stroke_color = None

    circle_repeat = _circle(270, 160, 60)
    circle_repeat.style.fill_color = GradientPaint(
        SweepGradient(
            cx=0.5,
            cy=0.5,
            start_angle=0.0,
            end_angle=1.5 * math.pi,
            stops=stops,
        ),
        units="objectBoundingBox",
        spread_method=GradientSpreadMethod.REPEAT,
    )
    circle_repeat.style.stroke_color = None

    circle_reflect = _circle(420, 160, 60)
    circle_reflect.style.fill_color = GradientPaint(
        SweepGradient(
            cx=0.5,
            cy=0.5,
            start_angle=0.0,
            end_angle=1.5 * math.pi,
            stops=stops,
        ),
        units="objectBoundingBox",
        spread_method=GradientSpreadMethod.REFLECT,
    )
    circle_reflect.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(circle_pad)
        dc.add_item(circle_repeat)
        dc.add_item(circle_reflect)

    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / "gradient_sweep_types.pdf",
        tmp_path,
    )


def test_gradient_alpha_variance(tmp_path: Path):
    pdf = _new_pdf()

    gradient = LinearGradient(
        0,
        0,
        1,
        0,
        colors=[DeviceRGB(r=0, g=0, b=0, a=1), DeviceRGB(r=0, g=0, b=0, a=0)],
    )
    path1 = _rect(40, 40, 100, 40)
    path1.style.fill_color = GradientPaint(gradient, units="objectBoundingBox")

    path2 = _rect(160, 40, 100, 40)

    # Different gradient transform but same LinearGradient object
    path2.style.fill_color = GradientPaint(
        gradient,
        units="objectBoundingBox",
        gradient_transform=Transform.translation(-0.5, -0.5)
        .rotate(math.radians(180))
        .translate(0.5, 0.5),
    )

    path3 = _circle(130, 200, 50)
    gradient3 = RadialGradient(
        0.5,
        0.5,
        0.0,
        0.5,
        0.5,
        0.5,
        colors=[DeviceRGB(r=0, g=0, b=0, a=1), DeviceRGB(r=0, g=0, b=1, a=0)],
    )
    path3.style.fill_color = GradientPaint(gradient3, units="objectBoundingBox")
    path3.style.stroke_color = None

    with pdf.drawing_context() as dc:
        dc.add_item(path1)
        dc.add_item(path2)
        dc.add_item(path3)

    assert_pdf_equal(
        pdf, HERE / "generated_pdf" / "gradient_alpha_variance.pdf", tmp_path
    )



"""Tests related to positioning self.x and self.y on the page.
implemented:
* ln()
to be implemented:
* set_x()
* set_y()
* set_xy()
"""

from pathlib import Path
import math

import pytest

from fpdf import FPDF

# from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent

# pylint: disable=protected-access


def test_ln_before_text():  # Issue 937
    pdf = FPDF()
    pdf.add_page()

    # no text added yet, ln() uses default font_size
    pdf.x += 20
    prev_y = pdf.y
    pdf.ln()
    assert math.isclose(
        pdf.y - prev_y, pdf.font_size
    ), f"ln() before writing text didn't move y by font size ({pdf.y-prev_y} vs. {pdf.font_size})."
    assert (
        pdf.x == pdf.l_margin
    ), f"ln() before writing text didn't move x to l_margin ({pdf.x} vs. {pdf.l_margin})."

    # changed font size affects amount of movement
    pdf.set_font("helvetica", size=36)
    pdf.x += 30
    prev_y = pdf.y
    pdf.ln()
    assert math.isclose(
        pdf.y - prev_y, pdf.font_size
    ), f"ln() before writing text didn't move y by font size ({pdf.y-prev_y} vs. {pdf.font_size})."
    assert (
        pdf.x == pdf.l_margin
    ), f"ln() before writing text didn't move x to l_margin ({pdf.x} vs. {pdf.l_margin})."


def test_ln_by_h():
    pdf = FPDF()
    pdf.add_page()

    # ln() uses argument h.
    pdf.x += 20
    prev_y = pdf.y
    h = 15.0
    pdf.ln(h)
    assert math.isclose(
        pdf.y - prev_y, h
    ), f"ln() didn't move y by argument h ({pdf.y-prev_y} vs. {h})."
    assert (
        pdf.x == pdf.l_margin
    ), f"ln() with argument h didn't move x to l_margin ({pdf.x} vs. {pdf.l_margin})."

    # bad argument h
    with pytest.raises(TypeError):
        pdf.ln("hello")


def test_ln_by_lasth():
    pdf = FPDF()
    pdf.add_page()

    # ln() uses self._lasth after writing some text.
    pdf.set_font("helvetica", size=16)
    pdf.x += 20
    pdf.cell(text="something")
    h = pdf.font_size  # last *used* font height, no cell height given
    prev_y = pdf.y
    pdf.ln(h)
    assert math.isclose(
        pdf.y - prev_y, pdf._lasth
    ), f"ln() after cell() didn't move y by font height ({pdf.y-prev_y} vs. {h})."
    assert (
        pdf.x == pdf.l_margin
    ), f"ln() after cell() didn't move x to l_margin ({pdf.x} vs. {pdf.l_margin})."

    # once we have a self._lasth, changing the font size without writing anything should
    # not affect the movement.
    pdf.set_font("helvetica", size=36)
    prev_y = pdf.y
    pdf.ln(h)
    assert math.isclose(
        pdf.y - prev_y, pdf._lasth
    ), f"ln() after cell() didn't move y by font height ({pdf.y-prev_y} vs. {h})."
    assert (
        pdf.x == pdf.l_margin
    ), f"ln() after cell() didn't move x to l_margin ({pdf.x} vs. {pdf.l_margin})."

    # cell(h=x) should set _lasth to that
    h = 20
    pdf.x += 33
    pdf.cell(h=h, text="something")
    prev_y = pdf.y
    pdf.ln()
    assert math.isclose(
        pdf.y - prev_y, h
    ), f"ln() after cell() didn't move y by cell height ({pdf.y-prev_y} vs. {h})."
    assert (
        pdf.x == pdf.l_margin
    ), f"ln() after cell() didn't move x to l_margin ({pdf.x} vs. {pdf.l_margin})."


def test_ln_with_zero_h():
    pdf = FPDF()
    pdf.add_page()

    prev_y = pdf.y
    # Passing 0(zero) as h
    pdf.ln(0)
    assert math.isclose(
        pdf.y, prev_y
    ), f"ln() after call with 0 (zero) h parameter moves by y axis ({pdf.y} vs.{prev_y})"



#!/usr/bin/env python3
# Script that demonstrate the use of all 14 standard (core / builtin) PDF fonts
from fpdf import FPDF
from fpdf.fonts import CORE_FONTS

pdf = FPDF()
pdf.add_page()
for core_font in CORE_FONTS:
    style = ""
    if core_font.endswith("I"):
        core_font = core_font[:-1]
        style += "I"
    if core_font.endswith("B"):
        core_font = core_font[:-1]
        style += "B"
    text = core_font.capitalize()
    if "B" in style:
        text += " Bold"
    if "I" in style:
        text += " Italics"
    if core_font in ("symbol", "zapfdingbats"):
        pdf.set_font("Helvetica", size=24)
        pdf.cell(text=text + " : ", h=9)
    pdf.set_font(core_font, style=style, size=24)
    pdf.cell(text=text, h=9)
    pdf.ln()
pdf.output("core_fonts.pdf")



from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_setting_bad_date():
    doc = fpdf.FPDF()
    with pytest.raises(TypeError):
        doc.set_creation_date("i am not a date")


def test_setting_old_date(tmp_path):
    doc = fpdf.FPDF()
    doc.add_page()
    # 2017, April 18th, almost 7:09a
    doc.set_creation_date(datetime(2017, 4, 18, 7, 8, 55).replace(tzinfo=timezone.utc))
    assert_pdf_equal(doc, HERE / "setting_old_date.pdf", tmp_path, at_epoch=False)


def test_setting_non_utc_date(tmp_path):
    doc = fpdf.FPDF()
    doc.add_page()
    # 2017, April 18th, almost 7:09a
    doc.set_creation_date(
        datetime(2017, 4, 18, 7, 8, 55).replace(tzinfo=timezone(timedelta(hours=5)))
    )
    assert_pdf_equal(doc, HERE / "setting_non_utc_date.pdf", tmp_path, at_epoch=False)



from fpdf.outline import OutlineSection, build_outline_objs
from fpdf.syntax import DestinationXYZ, PDFString


def _serialize_outline(sections, first_object_id=1):
    n = first_object_id
    outline_objs = list(build_outline_objs(sections))
    for obj in outline_objs:
        obj.id = n
        if n > first_object_id:
            obj.dest.page_ref = f"{2 * obj.dest.page_number + 1} 0 R"
        n += 1
    output = "\n".join(obj.serialize() for obj in outline_objs)
    return output


def test_serialize_outline():
    sections = (
        OutlineSection(
            "Title 1", level=0, page_number=1, dest=DestinationXYZ(page=1, top=0)
        ),
        OutlineSection(
            "Subtitle 1.1", level=1, page_number=1, dest=DestinationXYZ(page=2, top=0)
        ),
        OutlineSection(
            "Title 2", level=0, page_number=1, dest=DestinationXYZ(page=3, top=0)
        ),
        OutlineSection(
            "Subtitle 2.1", level=1, page_number=1, dest=DestinationXYZ(page=4, top=0)
        ),
        OutlineSection(
            "Subtitle 2.2", level=1, page_number=1, dest=DestinationXYZ(page=5, top=0)
        ),
    )
    assert _serialize_outline(sections, first_object_id=6) == """\
6 0 obj
<<
/Count 2
/First 7 0 R
/Last 9 0 R
/Type /Outlines
>>
endobj
7 0 obj
<<
/Count 1
/Dest [3 0 R /XYZ 0 0 null]
/First 8 0 R
/Last 8 0 R
/Next 9 0 R
/Parent 6 0 R
/Title (Title 1)
>>
endobj
8 0 obj
<<
/Count 0
/Dest [5 0 R /XYZ 0 0 null]
/Parent 7 0 R
/Title (Subtitle 1.1)
>>
endobj
9 0 obj
<<
/Count 2
/Dest [7 0 R /XYZ 0 0 null]
/First 10 0 R
/Last 11 0 R
/Parent 6 0 R
/Prev 7 0 R
/Title (Title 2)
>>
endobj
10 0 obj
<<
/Count 0
/Dest [9 0 R /XYZ 0 0 null]
/Next 11 0 R
/Parent 9 0 R
/Title (Subtitle 2.1)
>>
endobj
11 0 obj
<<
/Count 0
/Dest [11 0 R /XYZ 0 0 null]
/Parent 9 0 R
/Prev 10 0 R
/Title (Subtitle 2.2)
>>
endobj"""


def test__serialize_outline_with_headless_hierarchy():  # issues 239
    sections = (
        OutlineSection(
            "?-1", level=1, page_number=2, dest=DestinationXYZ(page=2, top=0)
        ),
        OutlineSection(
            "?-1-1", level=2, page_number=2, dest=DestinationXYZ(page=2, top=0)
        ),
        OutlineSection("1", level=0, page_number=2, dest=DestinationXYZ(page=2, top=0)),
        OutlineSection(
            "1-1", level=1, page_number=2, dest=DestinationXYZ(page=2, top=0)
        ),
        OutlineSection(
            "1-1-1", level=2, page_number=2, dest=DestinationXYZ(page=2, top=0)
        ),
    )
    assert _serialize_outline(sections, first_object_id=6) == """\
6 0 obj
<<
/Count 2
/First 7 0 R
/Last 9 0 R
/Type /Outlines
>>
endobj
7 0 obj
<<
/Count 1
/Dest [5 0 R /XYZ 0 0 null]
/First 8 0 R
/Last 8 0 R
/Parent 6 0 R
/Title (?-1)
>>
endobj
8 0 obj
<<
/Count 0
/Dest [5 0 R /XYZ 0 0 null]
/Parent 7 0 R
/Title (?-1-1)
>>
endobj
9 0 obj
<<
/Count 1
/Dest [5 0 R /XYZ 0 0 null]
/First 10 0 R
/Last 10 0 R
/Parent 6 0 R
/Title (1)
>>
endobj
10 0 obj
<<
/Count 1
/Dest [5 0 R /XYZ 0 0 null]
/First 11 0 R
/Last 11 0 R
/Parent 9 0 R
/Title (1-1)
>>
endobj
11 0 obj
<<
/Count 0
/Dest [5 0 R /XYZ 0 0 null]
/Parent 10 0 R
/Title (1-1-1)
>>
endobj"""


def test_serialize_outline_with_hex_encoding():  # issue-458
    sections = (
        OutlineSection(
            "Title with non-ASCCI letters: éêè",
            level=0,
            page_number=1,
            dest=DestinationXYZ(page=1, top=0),
        ),
    )
    assert _serialize_outline(sections, first_object_id=1) == """\
1 0 obj
<<
/Count 1
/First 2 0 R
/Last 2 0 R
/Type /Outlines
>>
endobj
2 0 obj
<<
/Count 0
/Dest [3 0 R /XYZ 0 0 null]
/Parent 1 0 R
/Title <feff005400690074006c0065002000770069007400680020006e006f006e002d004100530043004300490020006c006500740074006500720073003a002000e900ea00e8>
>>
endobj"""


def test_serialize_outline_without_hex_encoding():  # issue-458
    PDFString.USE_HEX_ENCODING = False
    try:
        sections = (
            OutlineSection(
                "Title with non-ASCCI letters: éêè",
                level=0,
                page_number=1,
                dest=DestinationXYZ(page=1, top=0),
            ),
        )
        assert _serialize_outline(sections, first_object_id=1) == f"""\
1 0 obj
<<
/Count 1
/First 2 0 R
/Last 2 0 R
/Type /Outlines
>>
endobj
2 0 obj
<<
/Count 0
/Dest [3 0 R /XYZ 0 0 null]
/Parent 1 0 R
/Title ({'Title with non-ASCCI letters: éêè'.encode('UTF-16').decode('latin-1')})
>>
endobj"""
    finally:
        PDFString.USE_HEX_ENCODING = True  # restore default value



"""
Tests that ensure tables generated by fpdf2
can be extracted by well-know PDF tables extraction tools
"""

import sys
from pathlib import Path

import camelot
from pandas import DataFrame
from pandas.testing import assert_frame_equal
import pytest
import tabula

from test.table.test_table import TABLE_DATA

HERE = Path(__file__).resolve().parent
_TMP_DF = DataFrame(TABLE_DATA)
TABLE_DATA_AS_DF = DataFrame(_TMP_DF.values[1:], columns=_TMP_DF.iloc[0])


###############################################################################
################################### camelot ###################################
###############################################################################

if sys.platform not in ("cygwin", "win32", "darwin") and sys.version_info[:2] != (
    3,
    14,
):
    # Disabling tests as GhostScript is not installed in GitHub Actions pipeline under Windows or macOS

    @pytest.mark.parametrize("flavor", ("lattice", "stream"))
    @pytest.mark.parametrize(
        "filename",
        (
            "table_simple.pdf",
            "table_with_images.pdf",
            "table_with_images_and_img_fill_width.pdf",
            "table_with_headings_styled.pdf",
            "table_with_internal_layout.pdf",
        ),
    )
    def test_camelot_extract_simple_table(flavor, filename):
        _test_camelot_parse(HERE / filename, flavor, 4, 5)

    @pytest.mark.parametrize(
        "filename",
        (
            "table_with_minimal_layout.pdf",
            "table_with_single_top_line_layout.pdf",
        ),
    )
    def test_camelot_extract_table_ok_with_only_stream_flavor(filename):
        _test_camelot_parse(HERE / filename, "stream", 4, 5)

    @pytest.mark.parametrize(
        "filename",
        (
            "table_align.pdf",
            # "table_with_cell_fill.pdf",
        ),
    )
    def test_camelot_extract_two_tables(filename):
        _test_camelot_parse(HERE / filename, "lattice", 4, 5, table_count=2)

    @pytest.mark.xfail(
        reason="camelot does not successfully parse tables split on several pages"
    )
    @pytest.mark.parametrize("flavor", ("lattice", "stream"))
    def test_camelot_extract_two_pages_table(flavor):
        _test_camelot_parse(HERE / "table_with_multiline_cells.pdf", flavor, 2, 5)

    def _test_camelot_parse(pdf_path, flavor, col_count, row_count, table_count=1):
        tables = camelot.read_pdf(str(pdf_path), flavor=flavor)
        assert tables.n == table_count
        for table in tables:
            assert len(table.cols) == col_count
            assert len(table.rows) == row_count


###############################################################################
################################### tabula ####################################
###############################################################################


@pytest.mark.parametrize(
    "filename",
    (
        "table_simple.pdf",
        "table_with_headings_styled.pdf",
        # "table_with_internal_layout.pdf",  # tabula only parses the internal cells
        "table_with_minimal_layout.pdf",
        "table_with_single_top_line_layout.pdf",
    ),
)
def test_tabula_extract_simple_table(filename):
    dataframes = tabula.read_pdf(HERE / filename, pages="all")
    assert len(dataframes) == 1
    for df_raw in dataframes:
        df = df_raw.astype({"Age": str})
        assert_frame_equal(df, TABLE_DATA_AS_DF, check_names=False)


@pytest.mark.parametrize(
    "filename",
    (
        "table_align.pdf",
        "table_with_cell_fill.pdf",
    ),
)
def test_tabula_extract_two_tables(filename):
    dataframes = tabula.read_pdf(HERE / filename, pages="all")
    assert len(dataframes) == 2
    for df_raw in dataframes:
        df = df_raw.astype({"Age": str})
        assert_frame_equal(df, TABLE_DATA_AS_DF, check_names=False)


@pytest.mark.xfail(
    reason="tabula does not successfully parse tables split on several pages"
)
def test_tabula_extract_two_pages_table():
    dataframes = tabula.read_pdf(HERE / "table_with_multiline_cells.pdf", pages="all")
    assert len(dataframes) == 2
    for df in dataframes:
        _rows_count, cols_count = df.shape
        assert cols_count == 2



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

import pytest

HERE = Path(__file__).resolve().parent


def test_hyperlinks(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=24)

    pdf.y = 50
    pdf.cell(
        text="Link from FPDF.cell()",
        center=True,
        link="https://github.com/py-pdf/fpdf2",
    )

    pdf.set_xy(60, 100)
    pdf.write_html('<a href="https://py-pdf.github.io/fpdf2/">Link defined as HTML</a>')

    text = "Link set over an arbitrary area with FPDF.link()"
    x, y = 20, 150
    pdf.text(x=x, y=y, text=text)
    width = pdf.get_string_width(text)
    pdf.link(
        x=x,
        y=y - pdf.font_size,
        w=width,
        h=pdf.font_size,
        link="https://github.com/py-pdf/fpdf2/discussions",
    )

    assert_pdf_equal(pdf, HERE / "hyperlinks.pdf", tmp_path)


def test_link_alt_text(tmp_path):
    """
    It can be tested that the reference file for this test
    has the link description read out loud by the NVDA screen reader
    when opened with Adobe Acrobat Reader.
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=24)
    text = "py-pdf/fpdf2"
    pdf.text(x=80, y=150, text=text)
    width = pdf.get_string_width(text)
    line_height = 10
    pdf.link(
        x=80,
        y=150 - line_height,
        w=width,
        h=line_height,
        link="https://github.com/py-pdf/fpdf2",
        alt_text="GitHub repository of the fpdf2 library",
    )
    assert_pdf_equal(pdf, HERE / "link_alt_text.pdf", tmp_path)


def test_link_with_zoom_and_shift(tmp_path):
    pdf = FPDF()
    pdf.set_font("helvetica", size=24)
    pdf.add_page()
    link = pdf.add_link()
    pdf.set_link(link, page=2, x=pdf.epw / 4, y=pdf.epw / 3, zoom=4)
    pdf.set_xy(30, 50)
    pdf.cell(
        w=140,
        h=10,
        text="Link to 2nd page zoomed & shifted",
        border=1,
        align="C",
        link=link,
    )
    pdf.add_page()
    pdf.multi_cell(
        pdf.epw,
        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit,"
        " sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
        " Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."
        " Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur."
        " Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.",
    )
    # Drawing Adobe Reader viewport after clicking on the link,
    # with the right panel open. The initial zoom level does not matter.
    pdf.set_draw_color(r=255, g=0, b=0)
    pdf.rect(x=pdf.epw / 4, y=pdf.epw / 3, w=53.5, h=31)
    assert_pdf_equal(pdf, HERE / "link_with_zoom_and_shift.pdf", tmp_path)


def test_link_border(tmp_path):
    "Acrobat renders this border it but not Sumatra"
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=24)

    text = "https://github.com/py-pdf/fpdf2"
    height = 10
    x, y = 50, 150

    pdf.text(x=x, y=y, text=text)
    pdf.link(
        x=x,
        y=y - height,
        w=pdf.get_string_width(text),
        h=height,
        link="https://github.com/py-pdf/fpdf2",
        border_width=4,
    )

    assert_pdf_equal(pdf, HERE / "link_border.pdf", tmp_path)


def test_inserting_same_page_link_twice(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.link(
        x=pdf.l_margin,
        y=pdf.t_margin,
        w=pdf.epw,
        h=pdf.eph,
        link=pdf.add_link(page=2),
    )
    pdf.add_page()
    pdf.link(
        x=pdf.l_margin,
        y=pdf.t_margin,
        w=pdf.epw,
        h=pdf.eph,
        link=pdf.add_link(page=2),
    )
    assert_pdf_equal(pdf, HERE / "inserting_same_page_link_twice.pdf", tmp_path)


def test_inserting_link_to_non_existing_page():
    pdf = FPDF()
    pdf.add_page()
    pdf.link(
        x=pdf.l_margin,
        y=pdf.t_margin,
        w=pdf.epw,
        h=pdf.eph,
        link=pdf.add_link(page=2),
    )
    with pytest.raises(ValueError):
        pdf.output()


def test_inserting_link_with_no_page_number():
    pdf = FPDF()
    link = pdf.add_link()
    pdf.add_page()
    pdf.set_font("helvetica", size=12)
    with pytest.raises(ValueError):
        pdf.cell(text="Page 1", link=link)


def test_later_call_to_set_link(tmp_path):  # v2.6.1 bug spotted in discussion 729
    pdf = FPDF()
    pdf.set_font("helvetica")

    pdf.add_page()  # page 1
    link_to_section1 = pdf.add_link()
    pdf.cell(text="Section 1", link=link_to_section1)

    pdf.add_page()  # page 2
    pdf.set_link(link_to_section1, page=pdf.page)
    pdf.cell(text="Section 1: Bla bla bla")

    assert_pdf_equal(pdf, HERE / "later_call_to_set_link.pdf", tmp_path)


def test_link_to_other_document(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=24)

    pdf.set_xy(80, 50)
    pdf.cell(
        text="Link defined with FPDF.cell",
        border=1,
        align="C",
        link="links.pdf",
    )

    pdf.set_y(100)
    pdf.multi_cell(
        w=pdf.epw,
        text="Link defined with FPDF.multi_cell",
        border=1,
        align="C",
        link="links.pdf",
    )

    pdf.set_y(150)
    pdf.multi_cell(
        w=pdf.epw,
        text="Link defined with FPDF.multi_cell and markdown=True: [links.pdf](links.pdf)",
        border=1,
        align="C",
        markdown=True,
    )

    pdf.set_xy(60, 200)
    pdf.write_html('<a href="links.pdf">Link defined with FPDF.write_html</a>')

    text = "Link defined with FPDF.link"
    pdf.text(x=80, y=250, text=text)
    width = pdf.get_string_width(text)
    pdf.link(
        x=80,
        y=250 - pdf.font_size,
        w=width,
        h=pdf.h - 250,
        link="links.pdf",
    )

    assert_pdf_equal(pdf, HERE / "link_to_other_document.pdf", tmp_path)


def test_internal_links(tmp_path):
    pdf = FPDF()
    pdf.set_font("helvetica", size=24)

    pdf.add_page()
    pdf.y = 100
    pdf.cell(text="Page 1", center=True)

    pdf.add_page()

    pdf.set_xy(80, 50)
    pdf.cell(
        text="Link defined with FPDF.cell",
        border=1,
        align="C",
        link=pdf.add_link(page=1),
    )

    pdf.set_y(100)
    pdf.multi_cell(
        w=pdf.epw,
        text="Link defined with FPDF.multi_cell",
        border=1,
        align="C",
        link=pdf.add_link(page=1),
    )

    pdf.set_y(150)
    pdf.multi_cell(
        w=pdf.epw,
        text="Link defined with FPDF.multi_cell and markdown=True: [page 1](1)",
        border=1,
        align="C",
        markdown=True,
    )

    pdf.set_xy(60, 200)
    pdf.write_html('<a href="1">Link defined with FPDF.write_html</a>')

    text = "Link defined with FPDF.link"
    pdf.text(x=80, y=250, text=text)
    width = pdf.get_string_width(text)
    pdf.link(
        x=80,
        y=250 - pdf.font_size,
        w=width,
        h=pdf.h - 250,
        link=pdf.add_link(page=1),
    )

    assert_pdf_equal(pdf, HERE / "internal_links.pdf", tmp_path)



from fpdf import FPDF
from pystrich.datamatrix import DataMatrixEncoder, DataMatrixRenderer

# Define the properties of the barcode
positionX = 10
positionY = 10
width = 57
height = 57
cellsize = 5

# Prepare the datamatrix renderer that will be used to generate the pilimage
encoder = DataMatrixEncoder("[Text to be converted to a datamatrix barcode]")
encoder.width = width
encoder.height = height
renderer = DataMatrixRenderer(encoder.matrix, encoder.regions)

# Generate a pilimage and move it into the memory stream
img = renderer.get_pilimage(cellsize)

# Draw the barcode image into a PDF file
pdf = FPDF()
pdf.add_page()
pdf.image(img, positionX, positionY, width, height)
pdf.output("datamatrix.pdf")



from io import BytesIO
from urllib.parse import quote
from urllib.request import urlopen
from fpdf import FPDF

formula = "x^n + y^n = a/b"
height = 170
url = f"https://chart.googleapis.com/chart?cht=tx&chs={height}&chl={quote(formula)}"
with urlopen(url) as img_file:  # nosec B310
    img = BytesIO(img_file.read())

pdf = FPDF()
pdf.add_page()
pdf.image(img, w=30)
pdf.output("equation_google_charts.pdf")



from pathlib import Path

import pytest

from fpdf import FPDF
from fpdf.enums import DocumentCompliance
from fpdf.errors import PDFAComplianceError

HERE = Path(__file__).resolve().parent


def test_pdfa_2b_encryption():
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_2B)
    with pytest.raises(PDFAComplianceError) as error:
        pdf.set_encryption(owner_password="fpdf2")
    assert (
        str(error.value)
        == "Encryption is now allowed for documents compliant with PDF/A-2B"
    )


def test_pdfa_3b_encryption():
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_3B)
    with pytest.raises(PDFAComplianceError) as error:
        pdf.set_encryption(owner_password="fpdf2")
    assert (
        str(error.value)
        == "Encryption is now allowed for documents compliant with PDF/A-3B"
    )


def test_pdfa_2b_core_font():
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_2B)
    with pytest.raises(PDFAComplianceError) as error:
        pdf.set_font("helvetica", "", 12)
    assert (
        str(error.value)
        == "Usage of base fonts is now allowed for documents compliant with PDF/A-2B. Use add_font() to embed a font file"
    )


def test_pdfa_3b_core_font():
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_3B)
    with pytest.raises(PDFAComplianceError) as error:
        pdf.set_font("helvetica", "", 12)
    assert (
        str(error.value)
        == "Usage of base fonts is now allowed for documents compliant with PDF/A-3B. Use add_font() to embed a font file"
    )



"""
Code used both in pdfchecker.py & verapdf.py
"""

import json, os, sys
from collections import defaultdict
from multiprocessing import cpu_count, Pool

try:  # optional dependency to display a progress bar
    from tqdm import tqdm
except ImportError:
    tqdm = lambda _, total: _

MULTIPROC_PARALLELISM = True


def main(checker_name, analyze_pdf_file, argv, checks_details_url):
    if len(argv) != 2:
        print(argv, file=sys.stderr)
        print(
            f"Exactly one argument must be passed to {checker_name}.py", file=sys.stderr
        )
        sys.exit(2)
    elif argv[1] == "--print-aggregated-report":
        print_aggregated_report(checker_name, checks_details_url)
    elif argv[1] == "--process-all-test-pdf-files":
        process_all_test_pdf_files(checker_name, analyze_pdf_file)
    else:
        print(analyze_pdf_file(argv[1]))


def process_all_test_pdf_files(checker_name, analyze_pdf_file):
    pdf_filepaths = [
        entry.path
        for entry in scantree("test")
        if entry.is_file() and entry.name.endswith(".pdf")
    ]
    print(
        f"Starting parallel execution of {checker_name} on {len(pdf_filepaths)} PDF files with {cpu_count()} CPUs"
    )
    reports_per_pdf_filepath = {}
    if MULTIPROC_PARALLELISM:
        with Pool(cpu_count()) as pool:
            for pdf_filepath, report in tqdm(
                pool.imap_unordered(analyze_pdf_file, pdf_filepaths),
                total=len(pdf_filepaths),
            ):
                reports_per_pdf_filepath[pdf_filepath] = raise_on_error(report)
    else:
        for pdf_filepath in tqdm(pdf_filepaths, total=len(pdf_filepaths)):
            report = analyze_pdf_file(pdf_filepath)[1]
            reports_per_pdf_filepath[pdf_filepath] = raise_on_error(report)
    agg_report = aggregate(checker_name, reports_per_pdf_filepath)
    print(
        f"{checker_name} analysis succeeded - Failures:",
        len(agg_report["failures"]),
        "- Warnings:",
        len(agg_report["warnings"]),
        "- Errors:",
        len(agg_report["errors"]),
    )


def raise_on_error(report_or_error):
    """
    This error handling may seems strange or useless,
    but optionally returning an error from analyze_pdf_file()
    is good way to "bubble up" errors raised in this function
    that would otherwise be "hidden" by multiprocessing.Pool.imap_unordered
    """
    if isinstance(report_or_error, BaseException):
        print(
            "ERROR: re-running this script with MULTIPROC_PARALLELISM=False can improve the stacktrace"
        )
        raise report_or_error
    return report_or_error


def scantree(path):
    """Recursively yield DirEntry objects for given directory."""
    for entry in os.scandir(path):
        if entry.is_dir():
            yield from scantree(entry.path)
        else:
            yield entry


def aggregate(checker_name, reports_per_pdf_filepath):
    aggregated_report_filepath = f"{checker_name}-aggregated.json"
    agg_report = {
        "failures": defaultdict(list),
        "warnings": defaultdict(list),
        "errors": defaultdict(list),
    }
    try:
        with open(aggregated_report_filepath, encoding="utf8") as agg_file:
            prev_agg_report = json.load(agg_file)
        agg_report["failures"].update(prev_agg_report["failures"])
        agg_report["warnings"].update(prev_agg_report["warnings"])
        agg_report["errors"].update(prev_agg_report["errors"])
    except FileNotFoundError:
        print("Initializing a new JSON file for the aggregated report")
        report = list(reports_per_pdf_filepath.items())[0][1]
        if "version" in report:
            agg_report["version"] = report.pop("version")
    for pdf_filepath, report in reports_per_pdf_filepath.items():
        if report.get("failure"):
            agg_report["failures"][report["failure"]].append(pdf_filepath)
        elif report.get("warning"):
            agg_report["warnings"][report["warning"]].append(pdf_filepath)
        else:
            for error in report.get("errors", ()):
                agg_report["errors"][error].append(pdf_filepath)
    with open(aggregated_report_filepath, "w", encoding="utf8") as agg_file:
        json.dump(agg_report, agg_file, indent=4)
    return agg_report


def print_aggregated_report(checker_name, checks_details_url):
    aggregated_report_filepath = f"{checker_name}-aggregated.json"
    with open(aggregated_report_filepath, encoding="utf8") as agg_file:
        agg_report = json.load(agg_file)
    if "version" in agg_report:
        print(agg_report["version"])
    print("# AGGREGATED REPORT #")
    if agg_report["failures"]:
        print("Failures:")
        for failure, pdf_filepaths in sorted(agg_report["failures"].items()):
            print(
                f"* {failure}: x{len(pdf_filepaths)} - First 3 files: {', '.join(pdf_filepaths[:3])}"
            )
    if agg_report["warnings"]:
        print("Warnings:")
        for warning, pdf_filepaths in sorted(agg_report["warnings"].items()):
            print(
                f"* {warning}: x{len(pdf_filepaths)} - First 3 files: {', '.join(pdf_filepaths[:3])}"
            )
    if agg_report["errors"]:
        print("Errors:")
        for error, pdf_filepaths in sorted(
            sorted(agg_report["errors"].items(), key=lambda error: -len(error[1]))
        ):
            print(
                f"* {error}: x{len(pdf_filepaths)} - First 3 files: {', '.join(pdf_filepaths[:3])}"
            )
    print("\nDocumentation on the checks:", checks_details_url)
    fail_on_unexpected_check_failure(checker_name, agg_report)


def fail_on_unexpected_check_failure(checker_name, agg_report):
    "exit(1) if there is any non-passing & non-whitelisted error remaining"
    ignore_whitelist_filepath = f"scripts/{checker_name}-ignore.json"
    with open(ignore_whitelist_filepath, encoding="utf8") as ignore_file:
        ignored_error_codes = set(json.load(ignore_file)["errors"].keys())
    report_error_codes = set(agg_report["errors"].keys())
    non_whitelisted_errors = report_error_codes - ignored_error_codes
    if agg_report["failures"] or non_whitelisted_errors:
        print(
            "Non-whitelisted issues found:",
            ", ".join(
                sorted(agg_report["failures"].keys()) + sorted(non_whitelisted_errors)
            ),
        )
        sys.exit(1)
    non_matching_ignored_codes = ignored_error_codes - report_error_codes
    if non_matching_ignored_codes:
        print(
            "Those whitelisted error codes are not reported anymore:",
            ", ".join(non_matching_ignored_codes),
        )
        print(f"This is probably due to a {checker_name} version upgrade")
        print(
            "Those whitelisted error codes should me removed from:",
            ignore_whitelist_filepath,
        )
        sys.exit(1)
